/* Copyright (C) 1992-2014 David Sugar, Tycho Softworks.
 *
 * This file is part of worldvu.
 *
 * worldvu is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * worldvuvu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with worldvu.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
    Standard message management code

    0.4     David Sugar     Feb 19 1992

    VUPostRedraw posts the structure of a 'NULL' rec ptr to the redraw queue.
    Memory exception generated on large models.

    0.5     David Sugar     Feb 19 1992

    A new VUUpdate procedure has been created by removing drawing queue
    events from dispatch into a seperate function.

    0.6     David Sugar     Mar 30 1992

    set_action and sync_action are now used to force tick idling only
    after 1 tick of inactivity.  This helps vu keep up with fast typists
    better.

    Addition of alarm timer services (VUAlarm and VUAbsAlarm).  VU_IDLE
    now passes execution time in p1 and keyboard idle in p2.

    1.0b+   David Sugar     Apr 3 1993

    Addition of VUPending() service.
*/

#include <vu-config.h>
#include <worldvu/queue.h>

CONST   char    copyright[] = "VU Version 2.0 Copyright (c) 1995 Tycho Softworks.  All rights reserved.";

static  EVENTQ  events[MAXEVQ];
static  REDRAWQ redraws[MAXRDQ];

QUEUE   evq     ={MAXEVQ, 0, 0};       /* q prts */
static  QUEUE   redq    ={MAXRDQ, 0, 0};

static  int (*alarm_exec[MAX_TIMERS])()={TNULL,TNULL,TNULL,TNULL,TNULL,TNULL,TNULL,TNULL};
static  time_t  alarm_trigger[MAX_TIMERS];

int     shutdown_flag = 0;
int     shutdown_mode = 0;
int     redraw = 0;

const   unsigned    char    VU_release = 20;

void     VUAlarm(id,sec,proc)
int id,sec;
int (*proc)(TARGS);
{
    id %= MAX_TIMERS;
    time(&alarm_trigger[id]);
    alarm_trigger[id] += (time_t)sec;
    alarm_exec[id] = proc;
}

void     VUAbsAlarm(id,when,proc)
int id;
time_t  when;
int (*proc)(TARGS);
{
    id %= MAX_TIMERS;
    alarm_trigger[id] = when;
    alarm_exec[id] = proc;
}

void VUPostRedraw(mode, vu, rc)
EVENT   mode;
VU      vu;
RECT    *rc;
{
    register    int tail,t1;

    tail = t1 = redq.qtail;
    if(++tail >= redq.qsize)
        tail = 0;

    redq.qtail = tail;
    redraws[t1].redmode = mode;
    if(rc)              /* fix mem excp */
        redraws[t1].redrc = *rc;
    else
        memcpy(&redraws[t1].redrc, &viewport, sizeof(RECT));
    redraws[t1].redvu = vu;
}

void VUPostEvent(VU vu, EVENT event, PARM p1, PARM p2)
{
    register    int tail, t1;

    if(vu)
        if(VUGetAttr(vu, REMOVED))
            return;

    tail = t1 = evq.qtail;
    if(++tail >= evq.qsize)
        tail = 0;

    evq.qtail = tail;
    events[t1].evport = vu;
    events[t1].event = event;
    events[t1].ev_p1 = p1;
    events[t1].ev_p2 = p2;
}

BOOL     VUPending()
{
    if(redraw)
        return TRUE;
    if(evq.qtail != evq.qhead)
        return TRUE;
    if(redq.qtail != redq.qhead)
        return TRUE;
    if(VUTermPending())
        return TRUE;
    return FALSE;
}

void     VUUpdate()      /* primary event dispatcher */
{
    register    int head;
    register    VU vu;
    RECT        rc, fix, *drawn;
    int         skip;

    /* process any pending redraw requests/screen changes */

    while((head=redq.qhead) != redq.qtail) {
        switch(redraws[head].redmode) {
        case VU_REDRAW:
            vu=desktop;

            /* if beyond screen, ignore */

            memcpy(&fix, &redraws[head].redrc, sizeof(rc));
            if(fix.lt > viewport.rt || fix.tp > viewport.bt ||
               fix.rt < viewport.lt || fix.bt < viewport.tp)
                break;

            memcpy(&rc, &viewport, sizeof(rc));
            /* clip fix to screen */
            if(fix.tp > viewport.tp)
                viewport.tp = fix.tp;
            if(fix.bt < viewport.bt)
                viewport.bt = fix.bt;
            if(fix.lt > viewport.lt)
                viewport.lt = fix.lt;
            if(fix.rt < viewport.rt)
                viewport.rt = fix.rt;

            while(vu != NULL) {
                if(VUGetAttr(vu,VISIBLE)) {
                    skip = 0;
                    drawn = rcfix(vu);
                    if(drawn->tp > viewport.bt)
                        ++skip;
                    if(drawn->bt < viewport.tp)
                        ++skip;
                    if(drawn->rt < viewport.lt)
                        ++skip;
                    if(drawn->lt > viewport.rt)
                        ++skip;
                    if(!skip)
                        VUEvent(vu, VU_PAINT, 0, 0);
                }
                vu=vu->vuprev;
            }
            memcpy(&viewport, &rc, sizeof(rc));
            break;
        case VU_PAINT_SELECT:
            ++force_update;         /* allow completion */
        case VU_PAINT:
            VUEvent(redraws[head].redvu, redraws[head].redmode, 0, 0);
            break;
        default:
            break;
        }
        if(++redq.qhead >= redq.qsize)
            redq.qhead=0;
        ++redraw;
    }
    return;
}

void    VUFlush()
{
    VUUpdate();
    VUTermUpdate();
}

int  VUDispatch()        /* primary event dispatcher */
{
    register    int head;
    int         key,timer;
    time_t      now;
    register    VU      vu;

    /* process keyboard or timer activity during idle */

    if(0 != (key = VUTermKey())) {
        if(VUGetAttr(front, LOCKED)) {
            if(!(VUEvent(front, VU_KEYBOARD, key, insert_mode)))
                VUEvent(desktop, VU_KEYBOARD, key, insert_mode);
        }
        else
            VUActiveEvent(VU_KEYBOARD, key, insert_mode);
        time(&_event_time);
    }
    else {      /* if no key event */
        time(&now);
        for(timer = 0; timer < MAX_TIMERS; ++timer) {
            if(alarm_exec[timer]) {
                if(alarm_trigger[timer] <= now)
                    alarm_trigger[timer] += (time_t)(*alarm_exec[timer])(timer);
            }
        }
        VUAllEvent(VU_IDLE, (int)(now - _event_time), (int)( (now - _login_time) / 60l));
    }

    /* process all message events */

    while((head = evq.qhead) != evq.qtail) {
        if(NULL == (vu = events[head].evport))
            VUAllEvent(events[head].event, events[head].ev_p1, events[head].ev_p2);
        else if(!VUGetAttr(vu,REMOVED))
            VUEvent(vu,events[head].event, events[head].ev_p1, events[head].ev_p2);
        if(++evq.qhead >= evq.qsize)
            evq.qhead = 0;
    }

    /* perform screen update if changed */

    VUUpdate();

    if(redraw)
        redraw=VUTermUpdate();

    return !shutdown_flag;
}

BOOL    VUAllEvent(EVENT event, PARM p1, PARM p2)              /* dispatch event to port */
{
    register    VU  vu=front;
    while(vu) {
        if(VUEvent(vu, event, p1, p2))
            return TRUE;
        vu = vu->vunext;
    }
    return FALSE;
}

BOOL    VUActiveEvent(EVENT event, PARM p1, PARM p2)                 /* dispatch event to port */
{
    register    VU  vu=front;

    while(vu) {
        if((vu == front) || VUGetAttr(vu,ACTIVE))
            if(VUEvent(vu, event, p1, p2))
                return TRUE;
        vu = vu->vunext;
    }
    return FALSE;
}

