/* Copyright (C) 1992-2014 David Sugar, Tycho Softworks.
 *
 * This file is part of worldvu.
 *
 * worldvu is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * worldvuvu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with worldvu.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Line drawing stuff for VU.   Added Line Drawing control for
 * release 1.1.
 */

#include <vu-config.h>
#include <worldvu/vu.h>

static  void GRCell(CELL cell, int row, int col)
{
    CELL    old;
    if(row < rows && col < cols) {
        old=imageport[row * cols + col];
        if(!(old & GRAPHIC_CELL))
            old = (old & PALMASK_CELL) | GRAPHIC_CELL;
        old |= cell;
        VUDrawCell(old, row, col);
    }
}

void VUDrawHLine(int r, int c, int len)
{
    if(len == 1)
        GRCell(BOX_L | BOX_R, r, c);
    else {
        GRCell(BOX_R, r, c++);
        while(--len > 1)
            GRCell(BOX_L | BOX_R, r, c++);
        GRCell(BOX_L, r, c);
    }
}

void VUDrawVLine(int r, int c, int len)
{
    if(len == 1)
        GRCell(BOX_U | BOX_D, r, c);
    else {
        GRCell(BOX_D, r++, c);
        while(--len > 1)
            GRCell(BOX_U | BOX_D, r++, c);
        GRCell(BOX_U, r, c);
    }
}

void VUDrawBox(int row, int col, int len, int wid)
{
    VUDrawVLine(row, col, len);
    VUDrawVLine(row,col + wid - 1, len);
    VUDrawHLine(row, col, wid);
    VUDrawHLine(row + len - 1, col, wid);
}

void VUDrawGrid(int row, int col, int len, int wid, int hsize, int vsize)
{
    int i;
    for(i = 0; i < vsize; ++i)
        VUDrawHLine(row + i * len, col, hsize * wid);
    for(i = 0; i < hsize; ++i)
        VUDrawVLine(row, col + i * wid, vsize * len);
}

void VUFillBox(int row, int col, int len, int wid, PAL pal)
{
    int i, j;
    CELL    cell    = (pal << 8) | ' ';

    for(i = row; i < (row + len); ++i)
        for(j = col; j < (col + wid); ++j)
            VUDrawCell(cell, i, j);
}

DEFCONTROL(VUFillControl)
{
    int row, col;
    switch(event) {
    case VU_CREATE:
        return TRUE;
    case VU_PAINT:
        col = vu->vudrawn.lt + control->ccol;
        row = vu->vudrawn.tp + control->crow;
        VUFillBox(row, col,
            control->clen, control->cwid, vu->vupal + control->cstate);
        return TRUE;
    default:
        return FALSE;
    }
}

DEFCONTROL(VULineControl)
{
    int fw = 0, fh = 0;
    int row, col, len = 0;
    char    *p = (char *)control->ctext;
    CELL    cell = (CELL)((vu->vupal + 1) << 8);

    switch(event) {
    case VU_CREATE:
        return TRUE;
    case VU_PAINT:
        col=vu->vudrawn.lt + control->ccol;
        row=vu->vudrawn.tp + control->crow;
        while(*p) {
            if((*p < '0') || (*p > '9'))
                if(!len)
                    ++len;
            switch(*p) {
            case 'b':
            case 'B':
                row = vu->vudrawn.bt - len;
                break;
            case 'e':
            case 'E':
                col = vu->vudrawn.rt - len;
                break;
            case '0':
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
            case '8':
            case '9':
                len = len * 10 + (*p - '0');
                break;
            case 'p':
                cell = (CELL)((vu->vupal + len) << 8);
                break;
            case 'P':
                cell = (CELL)(len << 8);
                break;
            case 'w':
            case 'W':
                fw = len;
                break;
            case 'h':
            case 'H':
                fh = len;
                break;
            case 'f':
                VUFillBox(row, col, fh, fw, len + vu->vupal);
                break;
            case 'F':
                VUFillBox(row, col, fh, fw, len);
                break;
            case 'c':
            case 'C':
                while(len--)
                    VUDrawCell(cell,row,col + len);
                break;
            case 'u':
                VUDrawVLine(row - len, col, len + 1);
            case 'U':
                row -= len;
                break;
            case 'd':
                VUDrawVLine(row, col, len + 1);
            case 'D':
                row += len;
                break;
            case 'r':
                VUDrawHLine(row, col, len + 1);
            case 'R':
                col += len;
                break;
            case 'l':
                VUDrawHLine(row, col - len, len + 1);
            case 'L':
                col -= len;
                break;
            case 's':
            case 'S':
                while(len--)
                    VUDrawCell(cell | ' ', row, col++);
                break;
            case '$':
                while(len--)
                    VUDrawCell(cell | *(++p), row, col++);
                break;
            default:
                while(len--)
                    VUDrawCell(cell | *p, row, col++);
            }
            if((*p < '0') || (*p > '9'))
                len = 0;
            ++p;
        }
        return TRUE;
    default:
        return FALSE;
    }
}
