/* Copyright (C) 1992-2014 David Sugar, Tycho Softworks.
 *
 * This file is part of worldvu.
 *
 * worldvu is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * worldvuvu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with worldvu.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
    A multi-line text exit resource control box

    0.4     David Sugar     Feb 17 92

    Added 'VU_ACCEPT' event handling to perform default failure if no input
    is provided.  Inherited failure can occur if p2 is set to TRUE.  VU_ACCEPT
    is only called if a default cmsg had been set and if something derived
    from 'VUForm' is the calling procedure of the vu object in question.

    0.6     David Sugar     Mar 30 92

    New editing features added for multi-line edit support:  ^N insert line,
    ^Y delete current line.

            David Sugar     May 21 92

    Trailing blanks surpressed.

            David Sugar     June 10 92

    Edit now split with editln.  EditLn offers simpler handling of single
    line edit jobs.
*/

#include <vu-config.h>
#include <ctype.h>
#include <worldvu/vu.h>
#include <worldvu/keys.h>

static int linelen(char *p)
{
    int len = 0;

    while(*p && (*p != '\n')) {
        ++len;
        ++p;
    }
    return len;
}

static  void redrawline(CELL cell, char *p, int r, int c, int cnt, int mask)
{
    while(cnt--) {
        if(*p && (*p != '\n')) {
            if(mask)
                VUDrawCell(cell | mask, r, c++);
            else
                VUDrawCell(cell | (CELL)*p, r, c++);
            ++p;
        }
        else
            VUDrawCell(cell | 0x0020, r, c++);
    }
    ++redraw;
}

static  int seekpos(char *p, int *row, int *col)
{
    int     cc = 0, cr = 0, cp = 0;

    while(*p) {
        if((*row == cr) && (*col == cc))
            break;
        if(*p == '\n') {
            if(cr == *row)
                break;
            ++cr;
            cc = 0;
        }
        else
            ++cc;

        ++cp;
        ++p;
    }
    *row = cr;
    *col = cc;
    return cp;
}

DEFCONTROL(VUEditControl)
{
    char    *p=(char *)control->ctext, *op;
    char    *m, *n;
    int     row = control->crow + vu->vudrawn.tp;
    int     col = control->ccol + vu->vudrawn.lt;
    int     cr = 0, cc = 0, cp = 0;
    int     erow, ecol;
    int     mask;
    int     len;
    BOOL    active = FALSE;
    CELL    cell;

    if(vu->vucontrol==control)
        active=TRUE;

    cell = (CELL)((vu->vupal + 3) << 8);

    switch(event) {
    case VU_CREATE:
        control->cstate=strlen(p);
        return TRUE;
    case VU_ACCEPT:         /* form accept event */
        len=strlen(p);
        while(len) {
            if(p[len - 1] == ' ')
                --len;
            else
                break;
        }
        p[len]=0;

        if(!*p || p2 == TRUE) {     /* if empty edit buffer or above */
            while(cp < control->cstate) {
                if(*p=='\n') {
                    if(++cr >= control->clen) {
                        --cr;
                        break;
                    }
                    cc=0;
                }
                else
                    ++cc;
                ++p;
                ++cp;
            }
            vu->vucontrol = control;
            vu->vustatline = control->cmsg;
            vu->vucontrol = control;
            VUControlCursor(vu, control, cr, cc);
            return FALSE;
        }
        return TRUE;
    case VU_KEYBOARD:
        while(cp < control->cstate && *p) {
            if(*p == '\n') {
                if(++cr >= control->clen) {
                    --cr;
                    break;
                }
                cc=0;
            }
            else
                ++cc;
            ++p;
            ++cp;
        }
        if(!*p)
            control->cstate=cp;

        mask=0;
        if(p2 > 1) {
            mask=p2;
            p2=0;
        }

        switch(p1) {
        case KEY_HOME:
            if(!active)
                return FALSE;
            cr=0;
            cc=0;
            control->cstate=0;
            VUControlCursor(vu,control,cr,cc);
            return TRUE;
        case KEY_END:
            if(!active)
                return FALSE;
            while(*p) {
                if(*p=='\n') {
                    ++cr;
                    cc = 0;
                }
                else
                    ++cc;
                ++control->cstate;
                ++p;
            }
            VUControlCursor(vu,control,cr,cc);
            return TRUE;
        case KEY_UP:
            if(active) {
                if(cr == 0)
                    return FALSE;
                --cr;
                control->cstate=seekpos((char *)control->ctext, &cr, &cc);
                VUControlCursor(vu, control, cr, cc);
                return TRUE;
            }
        case KEY_DOWN:
            if(active) {
                if(cr == (control->clen - 1))
                    return FALSE;
                if(!*p)
                    return FALSE;
                ++cr;
                control->cstate=seekpos((char *)control->ctext, &cr, &cc);
                VUControlCursor(vu, control, cr, cc);
                return TRUE;
            }
        case KEY_LEFT:
            if(active) {
                if(cr==0 && cc == 0)
                    return FALSE;
                --control->cstate;
                if(cc > 0) {
                    VUControlCursor(vu, control, cr, --cc);
                    return TRUE;
                }
                break;
            }
        case KEY_RIGHT:
            if(active) {
                if((cr == 0) && (cc == (control->cwid -1)))
                    return FALSE;
                if(!*p)
                    return FALSE;

                ++control->cstate;
                if(*p != '\n') {
                    VUControlCursor(vu, control, cr, ++cc);
                    return TRUE;
                }
                else {
                    VUControlCursor(vu, control, ++cr, 0);
                    return TRUE;
                }
            }

        case KEY_ENTER:
            if(active) {
                while(*p) {
                    if(*p == '\n') {
                        ++cr;
                        break;
                    }
                    ++p;
                    ++control->cstate;
                }

                if(*p == '\n') {
                    if(cr >= control->clen)
                        return FALSE;
                    ++p;
                    ++control->cstate;
                }
                else if(++cr < control->clen) {
                    *p = '\n';
                    *(++p) = 0;
                    ++control->cstate;
                }
                else
                    return FALSE;

                VUControlCursor(vu, control, cr, 0);
                return TRUE;
            }
        case KEY_PAGEUP:
        case KEY_PAGEDOWN:
        case KEY_TAB:
            if(active)
                return FALSE;
            VUControlCursor(vu, control, cr, cc);
            vu->vucontrol = control;
            return TRUE;
        case 14:
            if(!active)
                return FALSE;
            if(control->clen == 1)
                return TRUE;
            m=p+1;
            while(*p != 0)
                ++p;
            n=p+1;
            while(p >= m)
                *(n--) = *(p--);
            *m = '\n';
            break;
        case 24:
            if(!active)
                return FALSE;
            control->ctext[0] = 0;
            control->cstate = 0;
            break;
        case 25:
            if(!active)
                return FALSE;
            m=p;
            if(*m == '\n' && m > ((char *)control->ctext)) {
                --control->cstate;
                --m;
            }
            while(*m != '\n' && m > ((char *)control->ctext)) {
                --control->cstate;
                --m;
            }
            while(*p && *p != '\n')
                ++p;
            if(*p == '\n')
                ++p;
            while(*p)
                *(m++) = *(p++);
            *m = 0;
            break;
        case 8:
        case KEY_RUBOUT:
            if(!active)
                return FALSE;
            if(cc == 0)
                return TRUE;
            op = --p;
            while(*p) {
                *p = *(p+1);
                ++p;
            }
            --control->cstate;
            --cc;
            --vu->vucurcol;
            redrawline(cell, op, row + cr, col + cc, control->cwid - cc, mask);
            return TRUE;
        case KEY_DELETE:
            if(!active)
                return FALSE;
            if(!*p || (*p == '\n'))
                return TRUE;
            op = p;
            while(*p) {
                *p = *(p+1);
                ++p;
            }
            redrawline(cell, op, row + cr, col + cc, control->cwid - cc, mask);
            return TRUE;
        default:
            if(!active)
                return FALSE;

            if((p1 < 32) || (p1 > 126) || !active)
                return FALSE;

            if((p1 == 32) && !control->cstate)
                return TRUE;

            if(*p) {
                if(p2 || *p == '\n') {
                    n=p;
                    if((cc + linelen(p)) >=control->cwid)
                        return TRUE;
                    op=p+strlen(p);
                    *(op+1)=0;
                    while(op != p) {
                        *op=*(op-1);
                        --op;
                    }
                }
                *p = (char)p1;
                redrawline(cell, p, row + cr, col + cc, control->cwid - cc, mask);
                ++control->cstate;
                ++vu->vucurcol;
                return TRUE;
            }
            else if(cc < control->cwid) {
                *p = (char)p1;
                *(++p) = 0;
                if(mask)
                    VUDrawCell(cell | mask, row + cr, col + cc);
                else
                    VUDrawCell(cell | p1, row + cr, col + cc);
                ++vu->vucurcol;
                ++control->cstate;
                ++redraw;
                if(++cc >=control->cwid) {
                    if(++cr >=control->clen)
                        VUPostEvent(vu, VU_KEYBOARD, 9, 0);
                }
                return TRUE;
            }
            else if(VUGetAttr(vu,WORDWRAP) && (cr + 1) < control->clen) {
                n=p;
                while(*p != ' ' && *p != '\n' && p > ((char *)control->ctext)) {
                    --cc;
                    --p;
                }
                if(*p == ' ') {
                    ++control->cstate;
                    *n = (char)p1;
                    *(++n) = 0;
                    *p = '\n';
                }
                break;
            }
            else
                return TRUE;
        }

        ++redraw;
        cr = cc = cp = 0;
    case VU_PAINT:
        mask = 0;
        if(p2 > 1) {
            mask = p2;
            p2 = 0;
        }

        p=(char *)control->ctext;
        row=control->crow + vu->vudrawn.tp;
        col=control->ccol + vu->vudrawn.lt;
        erow = control->crow + vu->vudrawn.tp + control->clen;
        ecol = control->ccol + vu->vudrawn.lt + control->cwid;

        while(*p) {
            if(*p=='\n') {
                while(col < ecol)
                    VUDrawCell(cell | 0x020, row, col++);
                if(++row >= erow)       /* if past last line, stop */
                    return TRUE;
                col = control->ccol + vu->vudrawn.lt;
                if(cp < control->cstate) {
                    ++cr;
                    cc = 0;
                }
            }
            else if(col <ecol) {
                if(cp < control->cstate)
                    ++cc;
                if(mask)
                    VUDrawCell(mask | cell, row, col++);
                else
                    VUDrawCell((CELL) *p | cell, row, col++);
            }
            ++cp;
            ++p;
        }
        if(cp < control->cstate)
            control->cstate = cp;
        while(row <erow && col < ecol)
            VUDrawCell(cell | 0x020, row, col++);
        while(++row < erow) {
            col = control->ccol + vu->vudrawn.lt;
            while(col < ecol)
                VUDrawCell(cell | 0x020, row, col++);
        }
        if(active) {
            VUControlCursor(vu, control, cr, cc);
        }
        return TRUE;
    default:
        break;
    }
    return TRUE;
}

