/* Copyright (C) 1992-2014 David Sugar, Tycho Softworks.
 *
 * This file is part of worldvu.
 *
 * worldvu is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * worldvuvu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with worldvu.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
    A text resource control box

    0.4     David Sugar     Feb 17 92

    Added 'VU_ACCEPT' event handling to perform default failure if no input
    is provided.  Inherited failure can occur if p2 is set to TRUE.  VU_ACCEPT
    is only called if a default cmsg had been set and if something derived
    from 'VUForm' is the calling procedure of the vu object in question.

    0.6     David Sugar     Mar 30 92

    New editing features added for multi-line edit support:  ^N insert line,
    ^Y delete current line.

            David Sugar     May 21 92

    Trailing blanks surpressed.

            David Sugar     June 10 92

    EDIT broken into edit and editln.  Editln offers benifit of reduced
    overhead for simple line editing applications.  EditLn now inherited
    for all custom field edits instead of 'edit'.
*/

#include <vu-config.h>
#include <ctype.h>
#include <worldvu/vu.h>
#include <worldvu/multi.h>
#include <worldvu/keys.h>

static  int linelen(char *p)
{
    int len=0;

    while(*p) {
        ++len;
        ++p;
    }
    return len;
}

static  void redrawline(CELL cell, char *p, int r, int c, int cnt, int mask)
{
    while(cnt--) {
        if(*p) {
            if(mask)
                VUDrawCell(cell | mask,r,c++);
            else
                VUDrawCell(cell | (CELL)*p,r,c++);
            ++p;
        }
        else
            VUDrawCell(cell | 0x0020,r,c++);
    }
    ++redraw;
}

DEFCONTROL(VUEditLnControl)
{
    char    *p=(char *)control->ctext,*op;
    int     row=control->crow + vu->vudrawn.tp;
    int     col=control->ccol + vu->vudrawn.lt;
    int     cc=0,cp=0;
    int     ecol;
    int     mask;
    int     len;
    BOOL    active=FALSE;
    CELL    cell;

    if(vu->vucontrol==control)
        active=TRUE;

    cell = (CELL)((vu->vupal + 3) << 8);

    switch(event) {
    case VU_CREATE:
        control->cstate=strlen(p);
        return TRUE;
    case VU_ACCEPT:         /* form accept event */
        len = strlen(p);
        while(len) {
            if(p[len - 1] == ' ')
                --len;
            else
                break;
        }
        p[len] = 0;

        if(!*p || p2 == TRUE) {     /* if empty edit buffer or above */
            while(cp < control->cstate) {
                ++cc;
                ++p;
                ++cp;
            }
            vu->vucontrol=control;
            vu->vustatline=control->cmsg;
            vu->vucontrol=control;
            VUControlCursor(vu,control,0,cc);
            return FALSE;
        }
        return TRUE;
    case VU_KEYBOARD:
        while(cp < control->cstate && *p) {
            ++cc;
            ++p;
            ++cp;
        }
        if(!*p)
            control->cstate=cp;

        mask=0;
        if(p2 > 1) {
            mask=p2;
            p2=0;
        }

        switch(p1) {
        case KEY_HOME:
            if(!active)
                return FALSE;
            cc=0;
            control->cstate=0;
            VUControlCursor(vu,control,0,cc);
            return TRUE;
        case KEY_END:
            if(!active)
                return FALSE;
            while(*p) {
                ++cc;
                ++control->cstate;
                ++p;
            }
            VUControlCursor(vu,control,0,cc);
            return TRUE;
        case KEY_ENTER:
        case KEY_DOWN:
        case KEY_UP:
            if(active)
                return FALSE;
        case KEY_LEFT:
            if(active) {
                if(cc == 0)
                    return FALSE;
                --control->cstate;
                VUControlCursor(vu,control,0,--cc);
                return TRUE;
            }
        case KEY_RIGHT:
            if(active) {
                if(!*p)
                    return FALSE;
                ++control->cstate;
                VUControlCursor(vu,control,0,++cc);
                return TRUE;
            }
        case KEY_PAGEUP:
        case KEY_PAGEDOWN:
        case KEY_TAB:
            if(active)
                return FALSE;
            VUControlCursor(vu,control,0,cc);
            vu->vucontrol=control;
            return TRUE;
        case 24:
        case 25:
            if(!active)
                return FALSE;
            control->ctext[0]=0;
            vu->vucurcol -= control->cstate;
            control->cstate=0;
            redrawline(cell,(char *)control->ctext,row,col,control->cwid,mask);
            return TRUE;
        case 8:
        case 0x7f:
            if(!active)
                return FALSE;
            if(cc==0)
                return TRUE;
            op=--p;
            while(*p) {
                *p=*(p+1);
                ++p;
            }
            --control->cstate;
            --cc;
            --vu->vucurcol;
            redrawline(cell, op, row, col + cc, control->cwid - cc, mask);
            return TRUE;
        case KEY_DELETE:
            if(!active)
                return FALSE;
            if(!*p)
                return TRUE;
            op=p;
            while(*p) {
                *p = *(p+1);
                ++p;
            }
            redrawline(cell, op, row, col+ cc, control->cwid - cc, mask);
            return TRUE;
        default:
            if(!active)
                return FALSE;

            if(p1 > 255)
                return FALSE;

            if(!VUValidText(p1))
                return FALSE;

            if((p1==32) && !control->cstate)
                return TRUE;

            if(*p) {
                if(p2) {
                    if((cc + linelen(p)) >=control->cwid)
                        return TRUE;
                    op = p + strlen(p);
                    *(op + 1) = 0;
                    while(op != p) {
                        *op = *(op - 1);
                        --op;
                    }
                }
                *p = (char)p1;
                redrawline(cell, p, row, col + cc, control->cwid - cc, mask);
                ++control->cstate;
                ++vu->vucurcol;
                return TRUE;
            }
            else if(cc < control->cwid) {
                *p = (char)p1;
                *(++p) = 0;
                if(mask)
                    VUDrawCell(cell | mask, row, col + cc);
                else
                    VUDrawCell(cell | p1, row, col + cc);
                ++vu->vucurcol;
                ++control->cstate;
                ++redraw;
                if(++cc >=control->cwid)
                    VUPostEvent(vu,VU_KEYBOARD,9,0);
                return TRUE;
            }
            else
                return TRUE;
        }

    case VU_PAINT:
        if(VUGetAttr(vu,EDITBAR) && !active)
            cell = (CELL)(vu->vupal << 8);

        mask=0;
        if(p2 > 1) {
            mask=p2;
            p2=0;
        }

        p=(char *)control->ctext;
        row=control->crow + vu->vudrawn.tp;
        col=control->ccol + vu->vudrawn.lt;
        ecol = control->ccol + vu->vudrawn.lt + control->cwid;

        while(*p) {
            if(col <ecol) {
                if(cp < control->cstate)
                    ++cc;
                if(mask)
                    VUDrawCell(mask | cell,row,col++);
                else
                    VUDrawCell((CELL) *p | cell,row,col++);
            }
            ++cp;
            ++p;
        }
        if(cp < control->cstate)
            control->cstate=cp;
        while(col < ecol)
            VUDrawCell(cell | 0x020,row,col++);
        if(active) {
            VUControlCursor(vu,control,0,cc);
        }
        return TRUE;
    case VU_ENTER:
    case VU_EXIT:
        if(VUGetAttr(vu,EDITBAR))
            (*(control->cproc))(vu,control,VU_PAINT,0,0);
        break;
    default:
        break;
    }
    return TRUE;
}

DEFCONTROL(VUEditPwControl)
{
    switch(event) {
    case VU_KEYBOARD:
        if(isalnum(p1) || ispunct(p1) || p1 < 32 || p1 > 126) {
            p2='X';
            break;
        }
        return FALSE;
    case VU_PAINT:
        p2='X';
        break;
    default:
        break;
    }
    return USECONTROL(VUEditLnControl);
}

DEFCONTROL(VUEditIdControl)
{
    switch(event) {
    case VU_KEYBOARD:
        if(isalnum(p1) || p1 < 32 || p1 > 126 || p1=='_')
            break;
        return FALSE;
    default:
        break;
    }
    return USECONTROL(VUEditLnControl);
}

#ifndef FNDEFINED
DEFCONTROL(VUEditFnControl)
{
    switch(event) {
    case VU_KEYBOARD:
        if(p1 == '.' && control->cstate)
            return USECONTROL(VUEditLnControl);
    default:
        break;
    }
    return USECONTROL(VUEditIdControl);
}
#endif

DEFCONTROL(VUEditNumControl)
{
    switch(event)
    {
    case VU_KEYBOARD:
        if(isdigit(p1) || p1=='.' || p1=='-' || p1=='+' || p1==',' || p1 < 32 || p1 > 126)
            break;
        if(isalpha(p1))
            return TRUE;
        return FALSE;
    default:
        break;
    }
    return USECONTROL(VUEditLnControl);
}
