/* Copyright (C) 1992-2014 David Sugar, Tycho Softworks.
 *
 * This file is part of worldvu.
 *
 * worldvu is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * worldvuvu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with worldvu.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
    The memo box shows a multi-line memopad of text.  A master
    list of pointers holds the base address of each starting
    line.  The arrow and page keys may be used to scroll through
    the memo text.  The memo text is always word wrapped.

    Pointers for the text list block may be limited by memory
    model.
*/

#include <vu-config.h>
#include <worldvu/vu.h>
#include <worldvu/keys.h>
#include <worldvu/memo.h>

#ifdef __SMALL__
#define __PTR_LIST    512
#endif

#ifndef __PTR_LIST
#define __PTR_LIST  2048
#endif

int VUBuildMemoList(VU vu, char **list, char *text, int max)
{
    int pos = 0, width;
    int cnt = 1;
    RECT *rc = &(vu->vudrawn);
    char *space = NULL;

    width = rc->rt - rc->lt  + 1;
    if(VUGetAttr(vu, HASBORDER))
        width -= 2;

    *(list++) = text;

    while(*text && cnt < max) {
        if(*text == '\n') {
            pos = 0;
            ++cnt;
            *(list++) = ++text;
            space = NULL;
        }
        else if(pos < width - 2) {
            if(*text == ' ')
                space = text;
            ++pos;
            ++text;
        }
        else {
            if(space)
                text = space;
            else
                --text;
            pos = 0;
            ++cnt;
            *(list++)=++text;
            space = NULL;
        }
    }
    *list = NULL;
    return cnt;
}

DEFPORT(VUMemoboxPort)
{
    MEMO memo = (MEMO)(vu->vuoptions);
    CELL arrow;
    int lines = vu->vudrawn.bt - vu->vudrawn.tp;

    if(VUGetAttr(vu, HASBORDER))
        --lines;

    switch(event) {
    case VU_CREATE:
        memo->memo_idx = (char **)malloc((__PTR_LIST + 1) * sizeof(char *));
        if(!memo->memo_idx)
            return FALSE;
        memo->memo_size = VUBuildMemoList(vu, memo->memo_idx, vu->vupaint, __PTR_LIST);
        memo->memo_base = 0;
        break;
    case VU_DESTROY:
        free(memo->memo_idx);
        break;
    case VU_PAINT:
        if(VUGetAttr(vu, HASBORDER)) {
            arrow = ((vu->vupal + 6) << 8) | GRAPHIC_CELL;
            if(speed > 12000 && vu != front)
                arrow += 0x0100;
            if(memo->memo_mark)
                VUDrawCell(arrow | ARROW_U, vu->vudrawn.tp + 1, vu->vudrawn.rt);
            if(memo->memo_mark + vu->vudrawn.bt - vu->vudrawn.tp < memo->memo_size)
                VUDrawCell(arrow | ARROW_D, vu->vudrawn.bt - 1, vu->vudrawn.rt);
        }
        break;
    case VU_KEYBOARD:
        if(vu != front)
            break;

        switch(p1) {
        case KEY_PAGEUP:
            while(memo->memo_base && lines--)
                --memo->memo_base;

            vu->vupaint=memo->memo_idx[memo->memo_base];
            VUPostRedraw(VU_PAINT, vu, NULL);
            return TRUE;
        case KEY_UP:
            if(memo->memo_base) {
                vu->vupaint=memo->memo_idx[--memo->memo_base];
                VUPostRedraw(VU_PAINT, vu, NULL);
            }
            return TRUE;
        case KEY_DOWN:
            if(memo->memo_base + 1 < memo->memo_size) {
                vu->vupaint=memo->memo_idx[++memo->memo_base];
                VUPostRedraw(VU_PAINT, vu, NULL);
            }
            return TRUE;
        case KEY_PAGEDOWN:
            while((memo->memo_base + 1) < memo->memo_size && lines--)
                ++memo->memo_base;

            vu->vupaint=memo->memo_idx[memo->memo_base];
            VUPostRedraw(VU_PAINT, vu, NULL);
            return TRUE;
        }
    default:
        break;
    }
    return USEPORT(VUMsgboxPort);
}

