/* Copyright (C) 1992-2014 David Sugar, Tycho Softworks.
 *
 * This file is part of worldvu.
 *
 * worldvu is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * worldvuvu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with worldvu.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
    Menu list control

    The parent control will create sub-window which will then have
    a list of all data values for the control.  The sub-window will
    access the parent process's private data area.  This is a seldom
    used method for a parent window to share private data with it's
    children.
*/

#include <vu-config.h>
#include <worldvu/vu.h>
#include <worldvu/keys.h>
#include <ctype.h>

static void paint_label(VU vu, CITEM *control, BOOL mode)
{
    int     row = vu->vudrawn.tp + control->crow;
    int     col = vu->vudrawn.lt + control->ccol;
    char    *p;
    CELL    active = 0x800;
    CELL    hitkey = 0x900;
    CELL    passive =0xa00;
    CELL    use;

    p = (char *)control->ctext;
    if(!control->crow) {
        if(vu == desktop) {
            passive = 0x500;
            hitkey = 0x400;
        }
        else {
            passive = (vu->vupal + 5) << 8;
            hitkey = (vu->vupal + 4) << 8;
        }
    }

    if(mode)
        use = active;
    else
        use = passive;

    while(*p) {
        if(*p == '~') {
            if(!mode)
                use = hitkey;
        }
        else {
            if(vu == desktop)
                VUDrawAbsCell((CELL) *p | use, row, col++);
            else
                VUDrawCell((CELL) *p | use, row, col++);
            if(mode)
                use=active;
            else
                use=passive;
        }
        ++p;
    }
    ++redraw;
}

/*  Private select stuff */

static DEFPORT(MenuPort)
{
    VU      parent = vu->vuparent;
    CITEM   *control = parent->vucontrol;
    SELECT  opts = (SELECT)vu->vuoptions;

    switch(event) {
    case VU_KEYBOARD:
        switch(p1) {
        case KEY_TAB:
        case KEY_LEFT:
        case KEY_RIGHT:
            VUPostEvent(vu, VU_DESTROY, 0, 0);
            VUPostEvent(parent, VU_KEYBOARD, p1, insert_mode);
            VUPostEvent(parent, VU_KEYBOARD, KEY_ENTER, insert_mode);
            return TRUE;
        case ' ':
        case KEY_ENTER:
            VUPostEvent(NULL, VU_MENU, control->ctag, opts->sel_mark + 1);
            return TRUE;
        }
        break;
    case VU_CREATE:
        ++control->cstate;
        vu->vupaint=control->cmsg;
        if(parent == desktop)
            vu->vupal = 16;
        else
            vu->vupal=parent->vupal;    /* get parent color */
        vu->vuresid = control->ctag;
        break;
    case VU_DESTROY:
        --control->cstate;
        VUPostEvent(NULL, VU_MENU, control->ctag, 0);
        return USEPORT(VUWindowPort);
    default:
        break;
    }
    return USEPORT(VUMenuPort);
}

static  DEFWINDOW(Menu)
    WIN_CLASS(MenuPort, (STRING)"")
    WIN_ATTR(1, 1, HASBORDER | HASSHADOW, 16, NULL)
ENDDEF

/*  Our Select List Control */

DEFCONTROL(VUMenuListControl)
{
    int     row = control->crow + vu->vudrawn.tp - viewport.tp;
    int     col = control->ccol + vu->vudrawn.lt;
    BOOL    active = FALSE;
    VU      child;
    char    *p = (char *)control->ctext;
    RECT    rc;

    if(vu->vucontrol == control)
        active=TRUE;

    switch(event) {
    case VU_KEYBOARD:

        /*  since our window may be marked as 'ACTIVE', we do not want
            to create multiple select list windows if one has already
            been created */

        if(control->cstate)
            return FALSE;

        if(row >= viewport.bt)
            row = row - control->clen - 1;

        switch(p1) {
        case KEY_UP:
        case KEY_DOWN:
        case KEY_ENTER:
        case ' ':
            if(active) {
                paint_label(vu, control, TRUE);
                Menu[0].cwid = control->cwid + 2;
                Menu[0].clen = control->clen + 2;
                child = VUCreateQ(Menu, NULL, row + 1, col - viewport.lt - 1, NULL);
                child->vuparent = vu;
                vu->vucontrol = control;
                return TRUE;
            }
            else
                return FALSE;

        case KEY_LEFT:
            if(active && control == vu->vucontrols)
                return TRUE;
            goto dokey;
        case KEY_RIGHT:
            if(active && !(control + 1)->cproc)
                return TRUE;
        case KEY_TAB:
dokey:
            if(active) {
                paint_label(vu,control,FALSE);
                rc.tp = rc.bt = vu->vudrawn.tp + control->crow;
                rc.lt = vu->vudrawn.lt + control->ccol;
                rc.rt = rc.lt + control->cwid - 1;
                VUPostRedraw(VU_REDRAW, NULL, &rc);
                return FALSE;
            }
            paint_label(vu, control, TRUE);
            VUControlCursor(vu, control, 0, 0);
            vu->vucontrol = control;
            return TRUE;
        default:
            if(active)
                return FALSE;

            while(*p && *p != '~')
                ++p;

            if(*p=='~') {
                if(VUKeyMatch(*(++p), p1)) {
                    paint_label(vu, control, TRUE);
                    VUControlCursor(vu, control, 0, 0);
                    vu->vucontrol = control;
                    Menu[0].cwid = control->cwid + 2;
                    Menu[0].clen = control->clen + 2;
                    child = VUCreateQ(Menu, NULL, row, col - viewport.lt - 1, NULL);
                    child->vuparent = vu;
                    return TRUE;
                }
            }
            return FALSE;
        }
        break;

    case VU_CREATE:
        return TRUE;
    case VU_ACCEPT:         /* form accept event */
        return TRUE;

    case VU_PAINT:
        paint_label(vu, control, active);
        return TRUE;

    default:
        break;
    }
    return TRUE;
}


