/* Copyright (C) 1992-2014 David Sugar, Tycho Softworks.
 *
 * This file is part of worldvu.
 *
 * worldvu is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * worldvuvu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with worldvu.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <vu-config.h>
#include <worldvu/vu.h>
#include <worldvu/multi.h>
#include <fcntl.h>

CONST char *nls_editsets8[] = {
    "US/ASCII",
#ifdef  BUILD_LOCALIZED
    "ISO/LATIN-1",
    "ISO/LATIN-1",
    "US/ASCII",
#endif
    NULL
};

CONST char *nls_editsets7[] = {
    "US/ASCII",
#ifdef  BUILD_LOCALIZED
    "US/ASCII",
    "US/ASCII",
    "US/ASCII",
#endif
    NULL
};

CONST char *nls_dow[] = {
    "sunday", "monday", "tuesday", "wednasday", "thursday", "friday", "saturday",
#ifdef  BUILD_LOCALIZED
    "domenica", "luned", "marted", "mercoled", "gioved", "venerd", "sabato",
    "luni", "marti", "miercuri", "joi", "vineri", "simbata", "duminica",
    "nedele", "pondeli", "utery", "streda", "ctvrtek", "patek", "sobota",
#endif
    NULL
};

CONST char *nls_months[] = {
    "january", "february", "march", "april", "may", "june", "july", "august", "september", "october", "november", "december",
#ifdef  BUILD_LOCALIZED
    "gennaio", "febrraio", "marzo", "aprile", "maggio", "guigno", "luglio", "agosto", "settembre", "ottobre", "novembre", "dicembre",
    "ianuarie", "februarie", "martie", "aprilie", "may", "iunie", "iulie", "august", "septembrie", "octombrie", "noiembrie", "decembrie",
    "leden", "unor", "brezen", "duben", "kveten", "cerven", "cervenec", "srpen", "zari", "rijen", "listopad", "prosinec",
#endif
    NULL
};

CONST char latin1set[32][2] = {
    {'`',  'a'},
    {'\'', 'a'},
    {'^',  'a'},
    {'~',  'a'},
    {'\"', 'a'},
    {'/',  'a'},
    {'a',  'e'},
    {'c',  ','},

    {'`',  'e'},
    {'\'', 'e'},
    {'^',  'e'},
    {'\"', 'e'},
    {'`',  'i'},
    {'\'', 'i'},
    {'^',  'i'},
    {'\"', 'i'},

    {'-',  'd'},
    {'~',  'n'},
    {'`',  'o'},
    {'\'', 'o'},
    {'^',  'o'},
    {'~',  'o'},
    {'\"', 'o'},
    {'.',  '.'},

    {'/',  'o'},
    {'`',  'u'},
    {'\'', 'u'},
    {'^',  'u'},
    {'\"', 'u'},
    {'\'', 'y'},
    {'\'', 'p'},
    {'\"', 'y'}
};

CONST ATTR desktop_colors[] =
{
    DARKBLUE | BLACK,
#ifdef  BUILD_LOCALIZED
    DARKGREEN | BLACK,
    DARKBLUE | BLACK,
    DARKBLUE | BLACK
#endif
};

static  unsigned char   collate[256], upper[256], lower[256];
static  unsigned char   valid[16];
static  char    **months;
static  char    **dow;
static  int     abbrev = 3;
static  char    truetab[16] = "ty123456789";
static  int     century = 1900;
static  long    time_offset = 0l;

extern  BOOL    codepage_loaded;

void    GetVUMulti(nr)
int     nr;
{
    int i;
    static  char    editset[32] = "VU_EDITSET=";

    for(i = 0; i < 16; ++i)
        valid[i] = 0;

    for(i = 0; i < 256; ++i) {
            collate[i] = (unsigned char)i;
            upper[i] = (unsigned char)i;
            lower[i] = (unsigned char)i;
    }

    /* standard ASCII case conversion defaults */

    for(i = 'A'; i <= 'Z'; ++i)
        lower[i] = (unsigned char)(i + 32);

    for(i = 'a'; i <= 'z'; ++i)
        upper[i] = (unsigned char)(i - 32);

    /* ISO latin-1 case conversion defaults */

    for(i = 224; i < 256; ++i)
        upper[i] = (unsigned char)(i - 32);

    for(i = 192; i < 224; ++i)
        lower[i] = (unsigned char)(i + 32);

    VUSetPalette(1, desktop_colors[nr], 0);
    months = (char **)&nls_months[nr * 12];
    dow = (char **)&nls_dow[nr * 7];

    if(codepage_loaded)
        strcat(editset, nls_editsets8[nr]);
    else
        strcat(editset, nls_editsets7[nr]);
}

unsigned char VUToLower(unsigned char c)
{
    return lower[c];
}

unsigned char VUToUpper(unsigned char c)
{
    return upper[c];
}

char *VUStrProper(STRING s)
{
    char *r = (char *)s;
    int flag = 1;

    while(*s) {
        if(ispunct(*s) || isspace(*s))
            ++flag;
        else if(flag) {
            *s = VUToUpper(*s);
            flag = 0;
        }
        else
            *s = VUToLower(*s);
        ++s;
    }
    return r;
}

char *VUStrLower(STRING s)
{
    char *r = (char *)s;

    while(*s) {
        *s = lower[*s];
        ++s;
    }
    return r;
}

char *VUStrUpper(STRING s)
{
    char *r = (char *)s;

    while(*s) {
        *s = upper[*s];
        ++s;
    }
    return r;
}

int VUStrncmp(STRING s1, STRING s2, int n)
{
    int diff = 0;

    while(n--) {
        diff = collate[*s1] - collate[*s2];
        if(diff)
            break;

        if(!*s1 && !*s2)
            break;

        ++s1;
        ++s2;
    }
    return diff;
}

int VUStrcmp(STRING s1, STRING s2)
{
    return  VUStrncmp(s1, s2, -1);
}

int VUStrnicmp(STRING s1, STRING s2, int n)
{
    int diff = 0;

    while(n--) {
        diff = collate[lower[*s1]] - collate[lower[*s2]];
        if(diff)
            break;

        if(!*s1 && !*s2)
            break;

        ++s1;
        ++s2;
    }
    return diff;
}

int VUStricmp(STRING s1, STRING s2)
{
    return VUStrnicmp(s1, s2, -1);
}

BOOL VUStrBoolTrue(STRING s)
{
    if(strchr(truetab, VUToLower(*s)))
        return TRUE;

    if((*s > '0') && (*s <= '9'))
        return TRUE;

    return FALSE;
}

BOOL VUValidText(unsigned char c)
{
    int     offset;
    int     mask = 0x01;

    if(c < 32 || c == 127)
        return FALSE;

    if(c < 128)
        return TRUE;

    offset = (c - 128) / 8;
    mask = mask >> (c % 8);

    return valid[offset] & mask;
}

static char *abbrev_code(char *s)
{
    static char abbr[8];

    strncpy(abbr, s, abbrev);
    abbr[abbrev] = 0;
    return abbr;
}

char *VUMultiDow(int d, BOOL abbr)
{
    if(!abbr)
        return dow[d % 7];
    else
        return abbrev_code(dow[d % 7]);
}

char    *VUMultiMonth(int m, BOOL abbr)
{
    if(!abbr)
        return months[m % 12];
    else
        return abbrev_code(months[m % 12]);
}

int VUMultiCentury(time_t *time)
{
    time_t  local = *time + time_offset;
    struct  tm  *dt = localtime(&local);

    return century + dt->tm_year;
}

struct  tm  *VULocalTime(time_t *time)
{
    time_t  local = *time + time_offset;
    struct  tm  *dt = localtime(&local);

    return dt;
}

static STRING _VUTextString(STRING s, unsigned char marker)
{
    static unsigned char buf[1024];

    int pos = 0;

    while(*s) {
        if(*s == marker) {
            if(*(++s) == marker)
                buf[pos++] = *(s++);
            else
                buf[pos++] = VU_MARKER_CELL;
            continue;
        }
        else if(*s == '\t') {
            buf[pos++] = ' ';
            while(pos % 8)
                buf[pos++] = ' ';
        }
        else if((*s > 223) && !codepage_loaded) {   /* latin-1 set */
            buf[pos++] = latin1set[*s - 224][0];
            buf[pos++] = latin1set[*s - 224][1];
        }
        else if((*s > 191) && !codepage_loaded) { /* latin-1 set */
            buf[pos++] = latin1set[*s - 192][0] - 32;
            buf[pos++] = latin1set[*s - 192][1];
        }
        else if(*s > 127)
            buf[pos++] = '.';
        else if(*s > 31 || *s == '\n')
            buf[pos++] = *s;
        else if(!marker) {
            buf[pos++] = '^';
            buf[pos++] = *s + '@';
        }
        else
            buf[pos++] = *s;
        ++s;
    }
    buf[pos] = 0;
    return buf;
}

STRING VUTextString(STRING s)
{
    return _VUTextString(s, 0);
}

STRING VUTextCtrlString(STRING s)
{
    return _VUTextString(s, '~');
}

int VUStrLength(STRING s)
{
    int len = 0;

    if(codepage_loaded)
        return strlen((char *)s);

    while(*s) {
        if(*s > 127)
            ++len;

        ++len;
        ++s;
    }
    return len;
}

