/* Copyright (C) 1992-2014 David Sugar, Tycho Softworks.
 *
 * This file is part of worldvu.
 *
 * worldvu is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * worldvuvu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with worldvu.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
    Primary generic vu management routines.

    0.4     David Sugar     Feb 17 92

    Will copy windef->cmsg to vu object status line.

    0.5     David Sugar     Feb 26 92

    Shutdown now no longer causes exit.
*/

#include <vu-config.h>
#include <worldvu/vu.h>
#include <worldvu/multi.h>
#include <worldvu/queue.h>
#include <worldvu/term.h>

#ifndef BUILD_LOCALIZED
#undef  NR_SUPPORT
#define NR_SUPPORT  0x01
#endif

time_t  _login_time,_event_time;

int     vu_national_date = NR_DATE_FORMAT;
int     vu_national_time = NR_TIME_FORMAT;
int     vu_national_abbrv = NR_ABBRV_FORMAT;
int     replacement_language = NR_DEFAULT;  /* default english */
unsigned long   replacement_support[] = {NR_SUPPORT};

VU      front   = NULL;     /* window at front list */
VU      back    = NULL;     /* window at back list */
VU      desktop = NULL;     /* desktop level */

static int blocker=0;   /* block unessisary refresh requests */

RECT *rcfix(VU vu)
{
    static RECT r;

    memcpy(&r, &vu->vudrawn, sizeof(r));
    if(color && VUGetAttr(vu, HASSHADOW)) {
        r.rt += 2;
        ++r.bt;
    }
    return &r;
}

void VUDestroy(VU vu)
{
    if(front != back) {
        if(front == vu) {
            front = vu->vunext;
            VUPostEvent(front, VU_FRONT, TRUE, 0);
            front->vuprev = NULL;
        }
        else if(back == vu) {
            back = vu->vuprev;
            back->vunext = NULL;
        }
        else {
            vu->vuprev->vunext = vu->vunext;
            vu->vunext->vuprev = vu->vuprev;
        }
    }
    else
        front = back = NULL;

    if(VUGetAttr(vu, VISIBLE) && !blocker)
        VUPostRedraw(VU_REDRAW, NULL, rcfix(vu));

    if(vu->vuimage != NULL)
        free(vu->vuimage);

    VUSetAttr(vu, REMOVED);
    free(vu);

    if(vu == desktop)
        desktop=NULL;
}

void VUChangeRect(VU vu, RECT *rc)
{
    if(VUGetAttr(vu, VISIBLE)) {
        VUPostRedraw(VU_REDRAW, NULL, rcfix(vu));
        ++force_update;
    }

    memcpy(&vu->vudrawn, rc, sizeof(RECT));
    if(VUGetAttr(vu ,VISIBLE)) {
        ++force_update;
        VUPostRedraw(VU_PAINT,vu,rc);
    }
}

void VUStartup()
{
    int     nr = NR_DEFAULT;
    char    *p;
    char    buf[65];
    FILE    *fp;
    char    national[65];
    char    *lang = "default", *region = "default";
    char    lbuf[128];
    static  char    vu_national[65] = "VU_NATIONAL=";
    char    *env;

#ifdef  BUILD_LOCALIZED
    env = getenv("VU_NATIONAL");
    if(env)
        if(!*env)
            env = NULL;

    if(!env) {
        env = getenv("LANG");
        if(env)
            if(!*env)
                env = NULL;
        if(env) {
            strcpy(buf, env);
            lang= strtok(buf, "._");
            strcpy(national, env);
            region = strtok(national, ".");
        }
        else
            lang = region = "default";

        fp = fopen(PREFIX_CONFDIR "/multi.db", "r");
        env = NULL;
        while(fp) {
            fgets(lbuf, sizeof(lbuf) - 1, fp);
            if(feof(fp))
                break;

            p = strtok(lbuf, " \t\n");
            if(strcmp(p, region))
                if(strcmp(p, lang))
                    continue;

            env = VUTrimString(strtok(NULL, ""), " \t", " \t\n");
            strcat(vu_national, env);
            putenv(vu_national);
            break;
        }
        if(fp)
            fclose(fp);
    }
    if(env) {
        strcpy(national, env);
        nr = atoi(strtok(national, " \t\n"));
        vu_national_date = atoi(strtok(NULL, " \t\n"));
        vu_national_time = atoi(strtok(NULL, " \t\n"));
        vu_national_abbrv = atoi(strtok(NULL, " \t\n"));
    }

    if(!(replacement_support[0] & (1 <<(replacement_language))))
        nr = NR_DEFAULT;
#endif

    replacement_language = nr;

    _event_time=time(&_login_time);
    if(!VUTermStartup()) {
        fputs("VU: unable to initialize\n",stderr);
        exit(-1);
    }
    GetVUCodepage(nr);
    GetVUMulti(nr);
}

void VUShutdown()
{
    ++blocker;
    while(front != NULL) {
        if(VUGetAttr(front, VISIBLE))
            VUEvent(front, VU_PAINT, 0, 0);
        VUEvent(front, VU_DESTROY, 0, TRUE);
    }

    VUTermUpdate();
    VUTermShutdown();
    shutdown_flag=0;
}

VU VUCreate(const CITEM *windef, STRING title, int row, int col, vuproc_t proc, ATTR attr, VU parent, PTR paint)
{
    register    VU vu;
    register    const CITEM *control = &windef[1];

    CITEM   *target;
    RECT    rc;
    int     controlcount = 1;
    int     wid, len;
    int     cnt;
    int     language_set = replacement_language;
    STRING  def_title = windef->ctext;
    STRING def_status = windef->cmsg;
    int nrflag = 0;

    while(control->cproc) {
        if(control->cproc == CINV) {
            language_set = control->ctag;
            if(language_set == replacement_language) {
                if(control->ctext)
                    def_title = control->ctext;
                if(control->cmsg)
                    def_status = control->cmsg;
            }
            ++control;
            if(!nrflag) {
                if(language_set == -1)
                    language_set = replacement_language;

                if(language_set == replacement_language)
                    ++nrflag;
            }
            continue;
        }
        if(language_set != replacement_language) {
            ++control;
            continue;
        }
        ++controlcount;
        ++control;
    }
    cnt=sizeof(CITEM) * controlcount;

    if(!(vu=(VU)malloc(sizeof(struct vuport)+cnt)))
        return NULL;

    control = &windef[1];
    nrflag = 0;
    language_set = replacement_language;

    target = vu->vucontrols;
    while(control->cproc) {
        if(control->cproc == CINV) {
            language_set = control->ctag;
            ++control;
            if(!nrflag) {
                if(language_set == -1)
                    language_set = replacement_language;

                if(language_set == replacement_language)
                    ++nrflag;
            }
            continue;
        }
        if(language_set != replacement_language) {
            ++control;
            continue;
        }
        memcpy(target++, control++, sizeof(CITEM));
    }
    memcpy(target, control, sizeof(CITEM));

    if(attr == 0xffff)
        attr = windef->crow;

    if(!title)
        title = (STRING)def_title;

    attr &= ~VISIBLE;           /* initialize invisible and at back */
    if(back)
        back->vunext = vu;
    else
        front = vu;
    vu->vunext = NULL;
    vu->vuprev = back;
    back = vu;

    if(windef->cwid == -1)
        wid = viewport.rt-viewport.lt + 1;
    else
        wid = windef->cwid;

    if(windef->clen == -1)
        len = viewport.bt-viewport.tp + 1;
    else
        len = windef->clen;

    if(col == -1)
        col = ((viewport.rt - viewport.lt + 1 - wid) / 2) + viewport.lt;
    else
        col += viewport.lt;

    if(row == -1)
        row = ((viewport.bt - viewport.tp + 1 - len) / 2) + viewport.tp;
    else
        row += viewport.tp;

    if(!proc)
        proc=(BOOL (*)(VUARGS))windef->cproc;

    rc.tp = row;
    rc.lt = col;
    rc.bt = row+len-1;
    rc.rt = col+wid-1;
    vu->vustatcolor = DARKWHITE | BLACK;
    vu->vustatline = (STRING)def_status;
    vu->vuparent = parent;
    if(!(vu->vutitle = title))
        vu->vutitle=(STRING)"";
    memcpy(&vu->vudrawn, &rc, sizeof(rc));
    memcpy(&vu->vuorigin, &rc, sizeof(rc));
    vu->vuattr = attr;
    vu->vupal = (PAL)windef->ccol;
    vu->vucurmode = 0;      /* no cursor */
    vu->vunotify = 0;       /* state passive */
    vu->vuscroll = 0;       /* full screen scroll */
    vu->vuproc = proc;
    vu->vuimage = NULL;
    vu->vupaint = paint;
    vu->vuresid = windef->ctag;
    vu->vucontrol = vu->vucontrols;

    if(VUEvent(vu, VU_CREATE, TRUE, 0))
        return vu;
    else {
        VUDestroy(vu);
        return NULL;
    }
}

VU VUCreateQ(const CITEM *windef,STRING title, int row, int col, PTR paint)
{
    return  VUCreate(windef, title, row, col, VUNULL, 0xffff, front, paint);
}

VU VUCreateQR(CITEM *windef, STRING title, int row, int col, char *text)
{
    int l, w;
    int center;
    register CITEM *control = windef;
    VU vu;

    VUTextSize(text, &l, &w);
    control->cwid += w;
    control->clen += l;
    center = control->cwid / 2;
    while((++control)->cproc) {
        control->crow += l;
        control->ccol += center;
    }

    vu=VUCreate(windef, title, row, col, VUNULL, 0xffff, front, text);
    control = windef;
    control->cwid -= w;
    control->clen -= l;
    while((++control)->cproc) {
        control->crow -= l;
        control->ccol -= center;
    }
    return vu;
}

void VUToFront(VU vu)
{
    VUSetAttr(vu, VISIBLE);

    VUEvent(front, VU_FRONT, FALSE, 0); /* tell old window no longer front */

    if(vu != front) {
        if(vu==back) {
            back = vu->vuprev;
            back->vunext = NULL;
        }
        else {
            vu->vuprev->vunext = vu->vunext;
            vu->vunext->vuprev = vu->vuprev;
        }
        front->vuprev = vu;
        vu->vunext = front;
        front = vu;
        vu->vuprev = NULL;
        VUEvent(front, VU_FRONT, TRUE, 0);  /* tell new window is now font */
    }

    VUPostRedraw(VU_PAINT, vu, &vu->vudrawn);
}

void VUToBack(VU vu)
{
    VUClrAttr(vu, VISIBLE);

    if(vu == front) {
        VUEvent(front, VU_FRONT, FALSE, 0); /* tell old window no longer front */
        if(front->vunext)
            VUEvent(front->vunext, VU_FRONT, TRUE, 0);
    }

    if(vu != back) {
        if(vu == front) {
            front = vu->vunext;
            front->vuprev = NULL;
        }
        else {
            vu->vuprev->vunext = vu->vunext;
            vu->vunext->vuprev = vu->vuprev;
        }
        back->vunext = vu;
        vu->vuprev = back;
        back = vu;
        vu->vunext = NULL;
    }
    VUPostRedraw(VU_REDRAW, NULL, rcfix(vu));
}

void VUDestroyGroup(VU vu)
{
    register VU next = front;

    while(next) {
        if(next->vuparent == vu)
            VUDestroyGroup(next);
        next = vu->vunext;
    }
    VUPostEvent(vu, VU_DESTROY, 0, 1);
}

