/* Copyright (C) 1992-2014 David Sugar, Tycho Softworks.
 *
 * This file is part of worldvu.
 *
 * worldvu is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * worldvuvu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with worldvu.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
    0.5     David Sugar     Feb 25 92

    The operation of the 'generic' select class has been standardized and
    made into a usable procedure in itself.  Select now assumes that
    vupaint points to an array of string pointers which can be
    painted into the select window and scrolled.  The last pointer must
    be NULL and is used to determine the total size of the list.  sel_left,
    when introduced, will create a left margin effect, and length will
    be computed.  sel_mark is now simply an index into the string list
    pointer table.  Custom master procedures can still be used if nessisary.

    0.6     David Sugar     Mar 11 92

    Actual implimentation of the select window code becomes more complete
    along with better handling of select paint events.  The paint event can
    now be trapped externally or performed with a default routine.

    0.6     David Sugar     May 11 92

    Notify now causes interior re-paint.  This makes updating of select
    data on changes made in a child window automatic.  Re-calculation of
    the select field marks are also performed if the size of the select
    list is changed.

*/

#include <vu-config.h>
#include <worldvu/vu.h>
#include <worldvu/keys.h>
#include <worldvu/multi.h>

static char *fix[1] = {NULL};

void VUDrawSeparator(VU vu,int line, PAL pal)
{
    register int i;
    CELL cell;

    cell = (pal << 8) | GRAPHIC_CELL;
    VUDrawCell(cell | BOX_R | BOX_U | BOX_D,vu->vudrawn.tp + line,vu->vudrawn.lt);
    VUDrawCell(cell | BOX_L | BOX_U | BOX_D,vu->vudrawn.tp + line,vu->vudrawn.rt);
    i = vu->vudrawn.lt + 1;
    while(i < vu->vudrawn.rt)
        VUDrawCell(cell | BOX_L | BOX_R, vu->vudrawn.tp + line,i++);
}

DEFPORT(VUSelectPort)
{
    register    SELECT opts =(SELECT)vu->vuoptions;
    register    CITEM *control;

    int     top,left,width;
    STRING  *list=(STRING *)vu->vupaint;
    STRING  *pos;
    int     lpos;
    STRING  ln;
    PAL     pal;
    int     rtn, i, t;
    RECT    rc;
    CELL    arrow;

    switch(event) {
    case VU_KEYBOARD:
        if(!opts->sel_size)
            break;
        switch(p1) {
        case KEY_HOME:
            opts->sel_mark = 0;
            opts->sel_base = 0;
            VUPostRedraw(VU_PAINT_SELECT, vu, NULL);
            return TRUE;
        case KEY_PAGEDOWN:
            if(opts->sel_base + opts->sel_list < opts->sel_size) {
                opts->sel_base += opts->sel_list;
                opts->sel_mark = opts->sel_base;
            }
            VUPostRedraw(VU_PAINT_SELECT, vu, NULL);
            return TRUE;
        case KEY_PAGEUP:
            opts->sel_base -= opts->sel_list;
            if(opts->sel_base < 0)
                opts->sel_base=0;
            opts->sel_mark = opts->sel_base;
            VUPostRedraw(VU_PAINT_SELECT, vu, NULL);
            return TRUE;
        case KEY_END:
            opts->sel_base = opts->sel_size - opts->sel_list;
            if(opts->sel_base < 0)
                opts->sel_base = 0;
            opts->sel_mark = opts->sel_size - 1;
            VUPostRedraw(VU_PAINT_SELECT, vu, NULL);
            return TRUE;
        case 11:
        case KEY_UP:
            if(opts->sel_mark > 0) {
                --opts->sel_mark;
                if(opts->sel_mark < opts->sel_base)
                    --opts->sel_base;
                VUPostRedraw(VU_PAINT_SELECT, vu, NULL);
            }
            return TRUE;
        case 10:
        case KEY_DOWN:
            if(opts->sel_mark < (opts->sel_size - 1)) {
                ++opts->sel_mark;
                if(opts->sel_base + opts->sel_list <= opts->sel_mark)
                    ++opts->sel_base;
                VUPostRedraw(VU_PAINT_SELECT, vu, NULL);
            }
            return TRUE;
        default:
            break;
        }
        break;
    case VU_CREATE:
        rtn = VUWindowPort(vu, event, p1, p2);
        opts->sel_list = vu->vudrawn.bt - vu->vudrawn.tp - 1;
        if(vu->vucontrols[0].cproc != CNULL)
            opts->sel_list = vu->vucontrols[0].crow - 2;

        opts->sel_mark = opts->sel_base=opts->sel_size=0;
        pos = list;

        if(pos) {
            while (*(pos++) != NULL)
                ++opts->sel_size;
        }
        else
            vu->vupaint = (char *)fix;

        control = vu->vucontrols;
        opts->sel_box = NULL;
        opts->sel_txt = NULL;
        i = 0;
        while(control->cproc != CNULL) {
            if(control->ctag == VU_SELECT) {
                if(control->cproc == VUBoxControl) {
                    opts->sel_box = control;
                    opts->sel_list = control->clen - 2;
                }
                else if(control->cproc == VUEditLnControl) {
                    opts->sel_txt = control;
                    ++i;
                }
                else if(control->cproc == VUTextControl && !i) {
                    opts->sel_txt=control;
                }
            }
            ++control;
        }
        return rtn;
    case VU_PAINT:
        if(!VUWindowPort(vu, event, p1, p2))
            return FALSE;
        if(opts->sel_box == NULL)
            if(opts->sel_list < (vu->vudrawn.bt - vu->vudrawn.tp - 1))
                VUDrawSeparator(vu, opts->sel_list, vu->vupal + 6);

        /* call paint select event immediately to complete painting of
           select object.  This is performed inside a call instead of a
           drop-thru so that if a custom port proc is being used, the
           proc can trap and seperately handle the paint select event.
        */

        return VUEvent(vu, VU_PAINT_SELECT, -1, -1);

    case VU_NOTIFY:
        if(vu != front)
            break;

        pos = list;
        opts->sel_size = 0;
        while (*(pos++) != NULL)
            ++opts->sel_size;

        if(!opts->sel_size) {
            opts->sel_mark = 0;
            opts->sel_base = 0;
        }
        else {
            if(opts->sel_mark >= opts->sel_size)
                opts->sel_mark = opts->sel_size - 1;
            if(opts->sel_base >= opts->sel_size)
                opts->sel_base = opts->sel_size - 1;
        }

    case VU_PAINT_SELECT:
        if((p1 != -1) || (p2 != -1))
            if(vu != front)
                return TRUE;

        if(opts->sel_box == NULL) {
            top = vu->vudrawn.tp + 1;
            width = vu->vudrawn.rt - vu->vudrawn.lt - 1;
            left = vu->vudrawn.lt + 1;
            if(VUGetAttr(vu, HASBORDER)) {
                arrow = (vu->vupal + 6) << 8 | GRAPHIC_CELL;
                if(speed > 12000 && vu != front)
                    arrow += 0x0100;
            }
            else
                arrow = 0;
        }
        else {
            top = opts->sel_box->crow + vu->vudrawn.tp + 1;
            width = opts->sel_box->cwid - 2;
            left = opts->sel_box->ccol + vu->vudrawn.lt + 1;
            arrow = (vu->vupal + 1) << 8 | GRAPHIC_CELL;
        }

        if(arrow) {
            if(opts->sel_base)
                VUDrawCell(arrow | ARROW_U, top, left + width);
            else
                VUDrawCell(arrow | BOX_U | BOX_D, top, left + width);
            if(opts->sel_base + opts->sel_list < opts->sel_size)
                VUDrawCell(arrow | ARROW_D, top + opts->sel_list - 1, left + width);
            else
                VUDrawCell(arrow | BOX_U | BOX_D, top + opts->sel_list - 1, left + width);
        }

        for(i=0; i < opts->sel_list; ++i) {
            lpos = opts->sel_base + i;
            pal = vu->vupal;
            if((ln = list[lpos]) == NULL) {
                rc.tp = top + i;
                rc.lt = left;
                rc.rt = left + width - 1;
                rc.bt = top + opts->sel_list - 1;
                VUImageFill(&rc, (pal << 8) | 0x20);
                i = opts->sel_list;
            }
            else if(!*ln) {
                t = 0;
                while(t < width)
                    VUDrawCell(BOX_L | BOX_R | GRAPHIC_CELL | (pal * 256), top + i, left + (t++));
            }
            else {
                ln = VUTextString(ln);
                if(!inverse_mode)
                    VUDrawString((STRING)"  ", top + i, left, 2, pal);
                if(lpos == opts->sel_mark) {
                    pal += 3;
                    if(!inverse_mode)
                        VUDrawString((STRING)"->", top + i, left, 2, pal);
                }
                if(inverse_mode)
                    VUDrawString(ln, top + i, left, width, pal);
                else
                    VUDrawString(ln, top + i, left + 2, width - 2, pal);
            }
        }
        if(event == VU_PAINT_SELECT)
            return TRUE;
    default:
        break;
    }
    return USEPORT(VUWindowPort);
}

DEFPORT(VUSelectListPort)
{
    register    SELECT sel = (SELECT)vu->vuoptions;

    STRING  *list = (STRING *)vu->vupaint;
    int     mark = sel->sel_mark;
    STRING  p;

    if(event == VU_KEYBOARD) {
        if(!USEPORT(VUSelectPort)) {
            if(vu != front)
                return FALSE;

            if(VUToLower(list[mark][0]) > VUToLower(p1))
                mark=0;
            else {
                p = list[mark + 1];
                if(VUToLower(list[mark][0]) == VUToLower(p1)) {
                    if(!p)
                        return TRUE;
                    if(VUToLower(*p) > VUToLower(p1))
                        return TRUE;
                }
                ++mark;
            }

            while(list[mark]) {
                if(VUToLower(list[mark][0]) == VUToLower(p1)) {
                    sel->sel_mark = mark;
                    if(sel->sel_mark - sel->sel_base >= sel->sel_list)
                        sel->sel_base = sel->sel_mark;
                    if(sel->sel_mark < sel->sel_base)
                        sel->sel_base=sel->sel_mark;
                    VUPostRedraw(VU_PAINT_SELECT, vu, NULL);
                    return TRUE;
                }
                ++mark;
            }
            return FALSE;
        }
        else
            return TRUE;
    }
    else
        return USEPORT(VUSelectPort);
}
