/* Copyright (C) 1992-2014 David Sugar, Tycho Softworks.
 *
 * This file is part of worldvu.
 *
 * worldvu is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * worldvuvu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with worldvu.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
    This module provides support functions for assisting in the
    operation of VU select lists.  Four functions have been
    provided to manipulate the lists used in VU select list
    windows:    list insert, list remove, list add, and list
    locate.  The first three modify the contents of the
    current select list, and hence adjust select list pointers
    which may require the object to be re-painted afterwards.

    These functions may be used by child windows to manipulate
    the contents of a parent select list window as needed.
*/

#include <vu-config.h>
#include <worldvu/vu.h>

#define CMPARGS char *s1, char *s2
typedef int (*cmproc_t)(CMPARGS);

/*  VUSelectAdd()

    Add item to end of select list.
*/

void VUSelectAdd(VU vu, char *item)
{
    SELECT  opts = (SELECT)(vu->vuoptions);
    char **list = (char **)(vu->vupaint);

    list[opts->sel_size] = item;
    ++opts->sel_size;
    list[opts->sel_size] = NULL;
}

/*  VUSelectInsert()

    Insert item into middle of select list.
*/

void VUSelectInsert(VU vu, int pos, char *item)
{
    SELECT opts = (SELECT)(vu->vuoptions);
    char **list = (char **)(vu->vupaint);
    int tail;

    if(pos > opts->sel_size || pos < 0)
        return;

    if(pos == opts->sel_size) {
        VUSelectAdd(vu, item);
        return;
    }

    tail = ++opts->sel_size;
    while(tail > pos) {
        list[tail] = list[tail - 1];
        --tail;
    }

    if(opts->sel_mark == tail && opts->sel_size > 1) {
        ++opts->sel_mark;
        if(opts->sel_base - opts->sel_mark <= opts->sel_size)
            ++opts->sel_base;
    }
    list[tail] = item;
}

/*  VUSelectRemove()

    Remove an item from the select list.
*/

void VUSelectRemove(VU vu, int pos)
{
    SELECT opts = (SELECT)(vu->vuoptions);
    char **list = (char **)(vu->vupaint);

    while(pos < opts->sel_size) {
        list[pos] = list[pos + 1];
        ++pos;
    }
    --opts->sel_size;
    if(opts->sel_mark > pos)
        --opts->sel_mark;
    if(opts->sel_base > opts->sel_mark)
        --opts->sel_base;
}

/*  VUSelectLocate()

    Locate the logical item position of an item, assuming the
    select list is presented in sorted order.
*/

int VUSelectLocate(VU vu, char *item, int (*cmp)(const char *s1, const char *s2))
{
    char **list = (char **)(vu->vupaint);
    int pos=0;

    while(list[pos]) {
        if((*cmp)(item,list[pos]) <= 0)
            return pos;
        ++pos;
    }
    return pos;
}

