/* Copyright (C) 1992-2014 David Sugar, Tycho Softworks.
 *
 * This file is part of worldvu.
 *
 * worldvu is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * worldvuvu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with worldvu.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <vu-config.h>
#include <fcntl.h>
#include <signal.h>
#include <worldvu/vu.h>
#include <worldvu/queue.h>
#include <worldvu/keys.h>
#include <worldvu/term.h>

#ifdef  QNX4
#define NAP_DELAY
#define TERMIO_SETUP
#include <sys/dev.h>
#include <sgtty.h>
#include <unistd.h>
#include <unix.h>
#include <termios.h>
#endif

#ifdef  SCO
#define SET_VMIN_VTIME
#define POLL_INPUT
#define NAP_DELAY
#define TERMIO_SETUP
#include <unistd.h>
#include <fcntl.h>
#include <poll.h>
#include <termios.h>
#include <termio.h>
#endif

#ifdef  COHERENT
#define POLL_INPUT
#define SGTTY_SETUP
#define NAP_DELAY
#include <sgtty.h>
#include <unistd.h>
#include <fcntl.h>
#include <poll.h>
#endif

#ifdef  HAVE_TERMIOS_H
#define INLINE  inline
#define TERMIO_SETUP
#define USLEEP_DELAY
#define SELECT_SETUP
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/select.h>
#endif

#ifdef  FIONREAD
#define TIOCQUERY   FIONREAD
#endif

#ifndef STDOUT
#define STDIN   0
#define STDOUT  1
#endif

#ifndef BUF
#define BUF 4096
#endif

#ifndef COLORSET
#define COLORSET    PREFIX_CONFDIR "/colorset"
#endif

#ifndef CODEPAGES
#define CODEPAGES   PREFIX_CONFDIR "/codepages-"
#endif

#ifndef DEFTERM
#define DEFTERM     "ansipc"
#endif
#define ATTRIBUTES  (BRIGHT | INVERSE | UNDERLINE | BLINK)

typedef struct {
    int keycode;
    char    *keytext;
}   KEY;

RECT        viewport;
CELL        *imageport;
int         rows, cols;
int         alert_mode = 2;
BOOL        color = TRUE;
BOOL        insert_mode = TRUE;
BOOL        inverse_mode = FALSE;
BOOL        console_mode = FALSE;
BOOL        keyboard_timer = TRUE;
BOOL        force_update = FALSE;
unsigned    speed = 0xffff;
ATTR        attrmask = COLORATTRMASK;
CELL        deskpattern = 0x0020;
BOOL        shadow_button = TRUE;
BOOL        inhibit_escape = FALSE;
BOOL        codepage_loaded = FALSE;
BOOL        region_mode = FALSE;
int         region_flag = 0;

static      BOOL        profile = FALSE;
static      unsigned    shadow_offset;
static      unsigned    char    codepage[256];
static      unsigned    char    keymap[256];
static      unsigned char   cbuf[BUF];
static      int     cpos = 0;
static      unsigned char *eolmap;
static      char    statline[128] = "";
static      int     cc = -1, cr = -1;
static      ATTR    attr = 0x07, attr1 = 0x07;
static      ATTR    clrmask = 0x0000;
static      ATTR    tstmask = COLORATTRMASK;
static      TERMDEF *term;
static      unsigned char       fcolor = 0, bcolor = 0;
static      int     offset;
static      unsigned char       ansicolors[] = {0, 4, 2, 6, 1, 5, 3, 7};
static      int     max_match;
static      int     alt_base = 0;
static      BOOL    graphics_mode = FALSE;
static      BOOL    cursor_visible = TRUE;
static      ATTR    attributes = ATTRIBUTES;
static      BOOL    inverse_status = FALSE;
static      int trim = 1;
static      int counter;

static void io_flush(void);
static void setattr(ATTR);
static void io_escape(char *s);
static void io_puts(char *s);
static void io_putc(int c);
static void io_putu(unsigned long code);

/*  OS interface section:

    OS specific operations include input query, timed input, and delay.
    Of these, timed input is usually the most complex.  io_startup() and
    io_shudown() are provided to configure tty settings.  These are usually
    set to 8n1 with no echo and raw mode.
*/

#ifdef  QNX4
#define io_query()  dev_ischars(STDIN)

static  int     io_getc(int flag)
{
    int timeout = 2;
    unsigned    char    buf;

    if(!flag) {
        if(!keyboard_timer)
            if(!io_query())
                return 0;

        timeout = 10;
    }

    if(!dev_read(STDIN, &buf, 1, 1, 0, timeout, 0, 0))
        return 0;
    if(!buf)
        buf= 0xff;
    return buf;
}
#endif

#ifdef  TERMIO_SETUP

static  struct  termios old_stty, new_stty;

#ifdef  SELECT_SETUP
static  fd_set  __in, __out, __other;

#ifndef TIOCQUERY
static  int io_query()
{
    static  struct  timeval timeout = {0l, 0l};

    FD_SET(STDIN, &__in);
    FD_SET(STDIN, &__other);

    if(select(STDIN + 1, &__in, &__out, &__other, &timeout) < 1)
        return 0;

    /* telnet disconnect */
    if(FD_ISSET(STDIN, &__other))
        exit(-1);


    return FD_ISSET(STDIN, &__in);
}
#endif

static  int io_getc(int flag)
{
    struct  timeval timeout;
    unsigned    char    buf = 0;

    FD_SET(STDIN, &__in);
    FD_SET(STDIN, &__other);

    timeout.tv_sec = 0l;
    timeout.tv_usec = 100000l * (long)trim;
    if(!flag) {
        if(keyboard_timer)
            timeout.tv_usec = 500000l;
        else
            timeout.tv_usec = 0l;
    }

    if(select(STDIN + 1, &__in, &__out, &__other, &timeout) > 0) {
        /* telnet disconnect */
        if(FD_ISSET(STDIN, &__other))
            exit(-1);

        if(!FD_ISSET(STDIN, &__in))
            return 0;

        read(STDIN, &buf, 1);
        if(!buf)
            buf = 0xff;
    }
    return buf;
}
#endif

static  void    io_startup()
{
    static unsigned speedtab[] =
        {0, 50, 75, 110, 134, 150, 200, 300, 600, 1200,
        1800, 2400, 4800, 9600, 19200, 38400, 57600};

#ifdef  SELECT_SETUP
    FD_ZERO(&__in);
    FD_ZERO(&__out);
    FD_ZERO(&__other);
    FD_SET(STDIN, &__in);
#endif
    tcgetattr(STDIN, &old_stty);
    memcpy(&new_stty, &old_stty, sizeof(new_stty));
    new_stty.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
    new_stty.c_oflag &= ~OPOST;
    new_stty.c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
    new_stty.c_cflag &= ~(CSIZE|PARENB);
    new_stty.c_cflag |= CS8;
#ifdef  SET_VMIN_VTIME
    new_stty.c_cc[VMIN] = 1;
    new_stty.c_cc[VTIME] = 0;
#endif
    tcsetattr(STDIN, TCSAFLUSH, &new_stty);
    speed = (unsigned)cfgetospeed(&old_stty);
    if(speed < 17)
        speed = speedtab[speed];
    else if(speed < 50)
        speed = 0xffff;
}

static  void    io_shutdown()
{
    tcsetattr(STDIN, TCSAFLUSH, &old_stty);
}

void    VUTermSuspend(BOOL cls)
{
    setattr(DARKWHITE | BLACK);
    if(cls) {
        cursor(0, 0);
        io_puts(term->term_eop);
    }
    else {
        cursor(rows - 1, 0);
        io_puts(term->term_eol);
    }
    if(cls != 2) {
        io_escape(term->term_cursor_show);
        io_escape(term->term_shutdown);
        io_flush();
        tcsetattr(STDIN, TCSANOW, &old_stty);
    }
    else
        io_flush();
}

void    VUTermContinue()
{
    tcsetattr(STDIN, TCSADRAIN, &new_stty);
    io_escape(term->term_startup);
    io_flush();
    cc = cr = -1;
    statline[0] = 0;
    VUForceRedraw();
}
#endif

#ifdef  POLL_INPUT

#ifndef TIOCQUERY
static  int io_query()
{
    static  struct  pollfd input = {STDIN, POLLIN, 0};

    if(1 != poll(&input, 1l, 0))
        return 0;
    else
        return 1;
}
#endif

static  int io_getc(int flag)
{
    static  struct  pollfd  input = {STDIN, POLLIN, 0};
    int     timeout = 50 * trim;
    unsigned char   buf;

    if(!flag) {
        if(!keyboard_timer)
            if(!io_query())
                return 0;

        timeout = 500;
    }

    if(1 != poll(&input, 1l, timeout)) {
        fflush(stdout);
        return 0;
    }

    read(STDIN, &buf, 1);
    if(!buf)
        buf = 0xff;
    return buf;
}

#endif

#ifdef  TIOCQUERY
static  int io_query()
{
    long    count = 0l;

    ioctl(STDIN, TIOCQUERY, &count);
    return (int)count;
}
#endif

#ifdef  SGTTY_SETUP

static  struct  sgttyb  old_stty, new_stty;
static int  speeds[15] = {50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800, 9600, 19200, 38400};

static  void    io_startup()
{
    ioctl(STDIN, TIOCGETP, &old_stty);
    memcpy(&new_stty, &old_stty, sizeof(new_stty));
    new_stty.sg_flags = RAW | CBREAK;
    ioctl(STDIN, TIOCSETN, &new_stty);
    if((old_stty.sg_ospeed > 0) && (old_stty.sg_ospeed < 16))
        speed = speeds[old_stty.sg_ospeed - 1];
}

static  void    io_shutdown()
{
    ioctl(STDIN, TIOCSETN, &old_stty);
}

void    VUTermSuspend(BOOL cls)
{
    setattr(DARKWHITE | BLACK);
    if(cls) {
        cursor(0, 0);
        io_puts(term->term_cls);
    }
    else {
        cursor(rows - 1, 0);
        io_puts(term->term_eol);
    }
    io_escape(term->term_cursor_show);
    io_escape(term->term_shutdown);
    io_flush();
    ioctl(STDIN, TIOCSETN, &old_stty);
}

void    VUTermContinue()
{
    ioctl(STDIN, TIOCSETN, &new_stty);
    io_escape(term->term_startup);
    io_flush();
    cr = cc = -1;
    statline[0] = 0;
    VUForceRedraw();
}
#endif

#ifdef  NAP_DELAY
void    sync_timer(tics)
int     tics;
{
    nap(50 * tics);
}
#endif

#ifdef  USLEEP_DELAY
void    sync_timer(tics)
int tics;
{
    usleep(tics * 50000l);
}
#endif

/*  Remaining code is non-OS specific!  */

#define chk_flush()     if(cpos > (BUF - 32))   io_flush()

static  void    io_flush()
{
    if(cpos)
        write(STDOUT, cbuf, cpos);
    counter += cpos;
    cpos = 0;
}

static void io_putu(c)
unsigned long c;
{
    if(c < 0x80) 
        io_putc(c & 0x7f);
    else if(c <= 0x000007ff) {
        io_putc((c >> 6) | 0xc0);
        io_putc((c & 0x3f) | 0x80);
    }
    else if(c <= 0x0000ffff) {
        io_putc((c >> 12) | 0xe0);
        io_putc((c >> 6) & 0x3f | 0x80);
        io_putc((c & 0x3f) | 0x80);
    }
}    

static void io_putc(c)
int     c;
{
    cbuf[cpos++] = (unsigned char)c;
    if(cpos > (BUF - 1))
        io_flush();
}

static void io_escape(char *s)
{
    if(*s == '[') {
        io_putc(27);
        io_putc('[');
        ++s;
    }
    else if(*s == '?') {
        io_putc(27);
        io_putc('[');
        io_putc('?');
        ++s;
    }

    while(*s) {
        io_putc(*(s++));
    }
}

static void io_puts(char *s)
{
    while(*s) {
        cbuf[cpos++] = *(s++);
        if(cpos > (BUF - 1))
            io_flush();
    }
}

static  unsigned char   set_color(code)
unsigned char       code;
{
    if(term->term_color_ansi)
        return ansicolors[code];
    else
        return code;
}

static  void    setattr(ATTR cattr)
{
    ATTR    cattr1;
    int     count = 0;
    unsigned    char    fore, back;
    unsigned    char    merge;
    BOOL        fixcolor = FALSE;

    if(cattr == attr1)
        return;

    attr1 = cattr;

    cattr &= attrmask;

    if(color) {
        if(cattr & 0x08)
            cattr = (cattr & ~0x08) | BRIGHT;
    }

    cattr1 = cattr;
    cattr &= tstmask;
    if(cattr != cattr1) {
        if(cattr & SUBUNDERLINE)
            cattr |= UNDERLINE;
        else if(cattr & SUBBRIGHT)
            cattr |= BRIGHT;
        else if(cattr & SUBINVERSE)
            cattr |= INVERSE;
    }
    cattr &= ~(SUBUNDERLINE | SUBINVERSE | SUBBRIGHT);
    if(cattr == attr)
        return;

    if(graphics_mode)
        if((attr & SPECIAL) != (cattr & SPECIAL)) {
            if(cattr & SPECIAL)
                io_puts(term->term_attr_graphics_on);
            else {
                if(term->term_attr_graphics_off[0])
                    io_puts(term->term_attr_graphics_off);
                else if(term->term_attr_graphics_on[0]) {
                    io_puts(term->term_attr_clear);
                    attr &= ~(clrmask | 0x77);
                    fixcolor = TRUE;
                    /* attr |= DARKWHITE | BLACK; */
                }
            }
        }

    if((attr & attributes) != (cattr & attributes)) {
        chk_flush();
        io_puts(term->term_attr_beg);
        if(term->term_attr_merge) {
            if(term->term_attr_beg[0]) {
                merge = term->term_attr_clear[0];
                if(cattr & BRIGHT)
                    merge |= term->term_attr_bright_on[0];
                if(cattr & INVERSE)
                    merge |= term->term_attr_inverse_on[0];
                if(cattr & BLINK)
                    merge |= term->term_attr_blink_on[0];
                if(cattr & UNDERLINE)
                    merge |= term->term_attr_underline_on[0];
                io_putc(merge);
            }
        }
        else {
            if(attr & clrmask) {
                if((attr & clrmask) != (cattr & clrmask)) {
                    attr &= ~(clrmask | 0x77);
                    /* attr |= DARKWHITE | BLACK; */
                    fixcolor = TRUE;
                    io_puts(term->term_attr_clear);
                    ++count;
                }
            }

            if((attr & BRIGHT) != (cattr & BRIGHT)) {
                if(count++)
                    io_puts(term->term_attr_sep);
                if(cattr & BRIGHT)
                    io_puts(term->term_attr_bright_on);
                else
                    io_puts(term->term_attr_bright_off);
            }

            if((attr & BLINK) != (cattr & BLINK)) {
                if(count++)
                    io_puts(term->term_attr_sep);
                if(cattr & BLINK)
                    io_puts(term->term_attr_blink_on);
                else
                    io_puts(term->term_attr_blink_off);
            }

            if((attr & INVERSE) != (cattr & INVERSE)) {
                if(count++)
                    io_puts(term->term_attr_sep);
                if(cattr & INVERSE)
                    io_puts(term->term_attr_inverse_on);
                else
                    io_puts(term->term_attr_inverse_off);
            }

            if((attr & UNDERLINE) != (cattr & UNDERLINE)) {
                if(count++)
                    io_puts(term->term_attr_sep);
                if(cattr & UNDERLINE)
                    io_puts(term->term_attr_underline_on);
                else
                    io_puts(term->term_attr_underline_off);
            }
            if(term->term_color_attr) {
                if(fixcolor || ((cattr & 0x07) != (attr & 0x07))) {
                    fore = (unsigned char)(set_color((unsigned char)(cattr & 0x07)) + fcolor);
                    if(count++)
                        io_puts(term->term_attr_sep);
                    io_puts(term->term_color_beg);
                    io_putc(fore);
                }
                if(fixcolor || ((cattr & 0x70) != (attr & 0x70))) {
                    back = (unsigned char)(set_color((unsigned char)((cattr & 0x70) >> 4)) + bcolor);
                    if(count++)
                        io_puts(term->term_attr_sep);
                    io_puts(term->term_color_sep);
                    io_putc(back);
                }
            }
        }
        io_puts(term->term_attr_end);
    }

    if(color && !term->term_color_attr) {
        chk_flush();
        if((attr & 0x77) != (cattr & 0x77)) {
            fore = (unsigned char)(set_color((unsigned char)(cattr & 0x07)) + fcolor);
            back = (unsigned char)(set_color((unsigned char)((cattr & 0x70) >> 4)) + bcolor);
            io_puts(term->term_color_beg);
            if(term->term_attr_merge) {
                if(term->term_color_order)
                    merge = (unsigned char)((unsigned char)(fore * 16) | (unsigned char)back | term->term_color_sep[0]);
                else
                    merge = (unsigned char)((unsigned char)(back * 16) | (unsigned char)fore | term->term_color_sep[0]);
                if(!term->term_attr_beg[0]) {
                    if(cattr & BRIGHT)
                        merge |= term->term_attr_bright_on[0];
                    if(cattr & BLINK)
                        merge |= term->term_attr_blink_on[0];
                }
                io_putc(merge);
            }
            else if(term->term_color_order) {
                io_putc(fore);
                io_puts(term->term_color_sep);
                io_putc(back);
            }
            else {
                io_putc(back);
                io_puts(term->term_color_sep);
                io_putc(fore);
            }
            io_puts(term->term_color_end);
        }
    }
    attr = cattr;
}

static  void    grcell(int cell)
{
    if(term->unicode_encoding) {
        switch(cell) {
        case 0x01:
        case 0x02:
        case 0x03:
            io_putu(0x2502);
            return;
        case 0x05:
            io_putu(0x2518);
            return;
        case 0x06:
            io_putu(0x2510);
            return;
        case 0x07:
            io_putu(0x2524);
            return;
        case 0x09:
            io_putu(0x2514);
            return;
        case 0x0a:
            io_putu(0x250c);
            return;
        case 0x0b:
            io_putu(0x251c);
            return;
        case 0x04:
        case 0x08:
        case 0x0c:
            io_putu(0x2500);
            return;
        case 0x0d:
            io_putu(0x2534);
            return;
        case 0x0e:
            io_putu(0x252c);
            return;
        case 0x0f:
            io_putu(0x253c);
            return;
        case BOX_FILL_TOP:
            io_putu(0x2580);
            return;
        case BOX_FILL_BOT:
            io_putu(0x2584);
            return;
        case BOX_SOLID:
            io_putu(0x2588);
            return;
        case BOX_SHADED:
            io_putu(0x2591);
            return;
        case ARROW_U:
            io_putu(0x2191);
            return;
        case ARROW_D:
            io_putu(0x2193);
            return;
        case ARROW_L:
            io_putu(0x2190);
            return;
        case ARROW_R:
            io_putu(0x2192);
            return;
        case CARD_CLUBS:
            io_putu(0x2663);
            return;
        case CARD_HEARTS:
            io_putu(0x2665);
            return;
        case CARD_DIAMONDS:
            io_putu(0x2666);
            return;
        case CARD_SPADES:
            io_putu(0x2660);
            return;
        default:
            break;
        }
    }

    if(cell < 16) {
        io_putc(term->char_boxset[cell]);
        return;
    }

    switch(cell) {
    case ARROW_U:
    case ARROW_D:
    case ARROW_L:
    case ARROW_R:
        io_putc(term->char_arrows[cell - ARROW_U]);
        return;

    case CARD_CLUBS:
    case CARD_HEARTS:
    case CARD_DIAMONDS:
    case CARD_SPADES:
        io_putc(term->char_cards[cell - CARD_CLUBS]);
        return;

    case BOX_FILL_TOP:
        io_putc(term->char_top);
        return;

    case BOX_FILL_BOT:
        io_putc(term->char_bot);
        return;

    case BOX_SOLID:
        io_putc(term->char_solid);
        return;

    case BOX_SHADED:
    default:
        io_putc(term->char_shade);
        return;
    }
}

int     VUTermWrite(char *addr, int len)
{
    io_flush();
    return write(STDOUT, addr, len);
}

int VUTermRead(char *addr, int len, char *term)
{
    int     pos = 0;
    unsigned    char    c;
    char    *base = addr;

    while(pos < len)
    {
        if(0 == (c = (unsigned char)io_getc(pos)))
            return pos;
        *(addr++) = c;

        if(term)
        {
            *addr = 0;
            if(strpbrk(base, term))
                return pos;
        }
        ++pos;
    }
    return pos;
}

void    VUTermBlank()
{
    io_puts(term->term_cls);
    io_flush();
    while(!VUTermKey())
        ;
    VUForceRedraw();
}

void    VUTermAlert(mode)
int     mode;
{
    if(!mode)
        mode = alert_mode;

    fixcursor();
    if(!strcmp(term->term_startup, "\033[11m"))
        io_puts("\033[10m");

    while(mode--)
        io_puts(term->term_beep);

    if(!strcmp(term->term_startup, "\033[11m"))
        io_puts("\033[11m");

    io_flush();
}

static  void    cursor_address(pos)
int     pos;
{
    char    buf[4];
    char    *p = buf;

    pos += offset;

    if(term->term_cursor_decimal)
    {
        if(pos > 99)
        {
            *(p++) = (char)(pos / 100 + '0');
            pos %= 100;
        }
        if(pos > 9)
        {
            *(p++) = (char)(pos / 10 + '0');
            pos %= 10;
        }

        *(p++) = (char)(pos + '0');
        *p = 0;
        io_puts(buf);
    }
    else
        io_putc((unsigned char)(pos & 0xff));
}

void    cursor(int r, int c)
{
    if((r == cr) && (c == cc))
        return;

    chk_flush();

    if((r == cr) && (c == (cc - 1)))
        io_puts(term->term_left);
    else if((r == (cr + 1)) && (c == cc))
        io_puts(term->term_down);
    else if((r == cr) && (c == (cc + 1)) && term->term_right[0])
        io_puts(term->term_right);
    else if((r == (cr - 1)) && (c == cc) && term->term_up[0])
        io_puts(term->term_up);
    else if(!c && term->term_cursor_vt[0])
    {
        io_puts(term->term_cursor_beg);
        cursor_address(r);
        io_puts(term->term_cursor_vt);
    }
    else
    {
        io_puts(term->term_cursor_beg);
        if(term->term_cursor_order)
        {
            cursor_address(r);
            io_puts(term->term_cursor_sep);
            cursor_address(c);
        }
        else
        {
            cursor_address(c);
            io_puts(term->term_cursor_sep);
            cursor_address(r);
        }
        io_puts(term->term_cursor_end);
    }
    cr = r;
    cc = c;
}

void    statln(STRING msg)
{
    cursor(rows - 1, 0);
    setattr(DARKWHITE | BLACK);

    chk_flush();
    io_puts(term->term_eol);
    io_puts((char *)msg);
    io_flush();
}

void    fixcursor()
{
    register    VU  vu = front;
    char    *stat;
    int     i;
    ATTR    attr;
    BOOL    hide = FALSE;
    int     r, c;

    if(vu == NULL)
        return;

    while(NULL == (stat=(char *)vu->vustatline))
    {
        if(NULL == (vu=vu->vunext))
            break;
    }

    if(NULL == vu)
    {
        attr=DARKWHITE | BLACK;
        stat="";
    }
    else
        attr=vu->vustatcolor;

    if(strcmp(stat, statline))
    {
        strcpy(statline, stat);
        cursor(rows - 1, 0);
        if(inverse_status)
        {
            if(attr == (DARKWHITE | BLACK))
                attr = DARKBLACK | WHITE;
            else
                attr |= WHITE;
        }
        setattr(attr);
        if(term->term_stat_line[0])
            io_puts(term->term_stat_line);
        else if(term->term_wrap_flag)
        {
            for(i = 0; i < cols; ++i)
                io_putc(' ');
        }
        else
            io_puts(term->term_eol);

        if(statline[0])
        {
            cursor(rows -1, (cols - strlen(statline))/2);
            io_puts(statline);
        }
    }

    if(!front->vucurmode)
        hide = TRUE;

    r=front->vucurrow + front->vudrawn.tp;
    c=front->vucurcol + front->vudrawn.lt;
    if((c > viewport.rt) || (c < viewport.lt) || (r < viewport.tp) || (r > viewport.bt))
        hide = TRUE;

    chk_flush();
    if(hide)
    {
        if(cursor_visible)
        {
            cursor_visible = FALSE;
            io_escape(term->term_cursor_hide);
        }
    }

    if(hide)
    {
        r = rows - 1;
        c = 0;
    }
    cursor(r, c);

    if(!hide && !cursor_visible)
    {
        io_escape(term->term_cursor_show);
        cursor_visible = TRUE;
    }
    io_flush();
}

static  int match(s1, s2, len)
char    *s1, *s2;
int     len;
{
    register    int pos = 0;

    if(!s2)
        return 1;

    while(pos < len)
    {
        if(*s1 != *s2)
        {
            if(max_match < pos)
                max_match = pos;

            return pos + 1;
        }
        ++s1;
        ++s2;
        ++pos;
    }
    if(!*s2)
        return 0;
    else
    {
        if(max_match < pos)
            max_match = pos;
        return pos + 1;
    }
}

int VUTermKey()
{
    static int keytab[] = {KEY_HOME, KEY_END, KEY_PAGEUP, KEY_PAGEDOWN, KEY_INSERT, KEY_DELETE, KEY_HELP, KEY_FIND};
    static int altkeys[26] = {30, 48, 46, 32, 18, 33, 34, 35, 23, 36, 37, 38, 50, 49, 24, 25, 16, 19, 31, 20, 22, 47, 17, 45, 21, 44};
    static KEY ansi[] = {{KEY_LEFT, "\033OD"}, {KEY_LEFT, "\033[1D"}, {KEY_RIGHT, "\033OC"}, {KEY_RIGHT, "\033[1C"}, {KEY_UP, "\033OA"}, {KEY_UP, "\033[1A"}, {KEY_DOWN, "\033OB"}, {KEY_DOWN, "\033[1B"}};
    static int move[] = {KEY_UP, KEY_DOWN, KEY_LEFT, KEY_RIGHT};

    unsigned char   kbuf[8];
    int     kpos = 0;
    int     i;
    int     code;

    max_match = 0;

    while(kpos < 8) {
        code = keymap[io_getc(kpos)];

        if(alt_base && !kpos) {
            if(code >= alt_base && code < (alt_base + 26))
                return ALTKEY | (((code - alt_base) & 0x1f) + 'A');
        }
        if(!code) {
            if(kpos == 1)
                return kbuf[0];

            i = strlen(term->key_alt_prefix);
            if(i && (i < kpos))
                if(!strncmp(term->key_alt_prefix, (char *)kbuf, i))
                    return ALTKEY | (kbuf[i] - alt_base);

            return 0;
        }

        kbuf[kpos] = (unsigned char)code;
        ++kpos;

        if(!match(kbuf, term->key_alt_prefix, kpos))
            return ALTKEY | ((io_getc(kpos) & 0x1f) + 0x40);

/*
        if(!match(kbuf, term->term_left, kpos))
            return KEY_LEFT;

        if(!match(kbuf, term->term_right, kpos))
            return KEY_RIGHT;

        if(!match(kbuf, term->term_up, kpos))
            return KEY_UP;

        if(!match(kbuf, term->term_down, kpos))
            return KEY_DOWN;
*/

        for(i = 0; i < 10; ++i)
            if(!match(kbuf, term->key_table[i], kpos))
                return KEY_F1 + i;

        if(!match(kbuf, term->key_table[10], kpos))
            return 0x01;

        if(!match(kbuf, term->key_table[11], kpos))
            return 0x03;

        for(i = 0; i < 4; ++i)
            if(!match(kbuf, term->key_move[i], kpos))
                return move[i];

        for(i = 12; i < 20; ++i)
            if(!match(kbuf, term->key_table[i], kpos)) {
                if(i == 16) {
                    if(insert_mode)
                        insert_mode = FALSE;
                    else
                        insert_mode = TRUE;
                    return 0;
                }
                return keytab[i - 12];
            }

        if(term->term_ansi_keypad)
            for(i = 0; i < (sizeof(ansi) / sizeof(KEY)); ++i) {
                if(!match(kbuf, ansi[i].keytext, kpos))
                    return ansi[i].keycode;
            }

        if(term->key_scancodes) {
            if((kpos == 1) && (code == 0xff) && !max_match)
                continue;

            if((kpos == 2) && (code == 224))
                continue;

            if(kbuf[0] == 0xff) {
                if(code > 58 && code < 69)
                    return KEY_F1 + code - 59;

                for(i = 0; i < 26; ++i)
                    if(altkeys[i] == code)
                        return ALTKEY + 'A' + i + 1;

                switch(code) {
                case 82:
                    if(insert_mode)
                        insert_mode = FALSE;
                    else
                        insert_mode = TRUE;
                    return 0;
                case 83:
                    return KEY_DELETE;
                case 71:
                    return KEY_HOME;
                case 79:
                    return KEY_END;
                case 73:
                    return KEY_PAGEUP;
                case 81:
                    return KEY_PAGEDOWN;
                case 72:
                    return KEY_UP;
                case 80:
                    return KEY_DOWN;
                case 75:
                    return KEY_LEFT;
                case 77:
                    return KEY_RIGHT;
                }
            }
        }

        if(kpos == 1 && !max_match)
            return kbuf[0];

        if(kpos > (max_match + 1))
            return 0;
    }
    return 0;
}

int VUTermUpdate()
{
    register    CELL    *mark;
    int c, r;
    BOOL        cflag;
    CELL        v;
    PAL     p1, p2;

    counter = 0;
    if(!force_update)
        if(VUTermPending())
            return 1;

    for(c = 0; c < shadow_offset; ++c) {
        if((imageport[c] & 0x40ff) != (imageport[c + shadow_offset] & 0x40ff))
            continue;

        p1 = (imageport[c] & PALMASK_CELL) >> 8;
        p2 = (imageport[c + shadow_offset] & PALMASK_CELL) >> 8;

        if(p1 == p2)
            continue;

        if((palette[p1].palattr & attrmask) != (palette[p2].palattr & attrmask))
            continue;

        imageport[c + shadow_offset] = imageport[c];
    }

    for(r = 0; r < rows - 1; ++r) {
        *(eolmap + r) = 0;

        mark = imageport + cols * r + cols;
        for(c = cols - 1;c > - 1;--c) {
            v = *(--mark);
            if((v != 0x0020) && (v != 0x0120)) {
                *(eolmap + r)=(unsigned char)(++c);
                break;
            }
        }
    }

    for(r = 0; r < rows - 1; ++r) {
        if(!force_update)
            if(VUTermPending()) {
                io_flush();
                return 1;
            }

        mark = imageport + r * cols;
        cflag = TRUE;

/*
        if(cursor_visible) {
            io_escape(term->term_cursor_hide);
                cursor_visible = FALSE;
        }
*/
        for(c = 0; c < cols; ++c) {
            if(*mark != *(mark + shadow_offset)) {
                *(mark + shadow_offset) = *mark;
                if(cflag) {
                    cursor(r, c);
                    cflag = FALSE;
                }

                if(eolmap[r] <= c) {
                    setattr(DARKWHITE | BLACK);
                    io_puts(term->term_eol);
                    while(c++ < cols) {
                        *(mark + shadow_offset) = deskpattern;
                        *(mark++) = deskpattern;
                    }
                    break;
                }

                if(*mark & GRAPHIC_CELL) {
                    setattr(palette[(*mark & 0x3f00) >> 8].palattr | SPECIAL);
                    grcell(*mark & 0xff);
                }
                else {
                    setattr(palette[(*mark & 0x3f00) >> 8].palattr);
                    io_putc(codepage[*mark & 0xff]);
                }
                ++cc;
            }
            else
                cflag = TRUE;
            ++mark;
        }
    }

    if(profile) {
        cursor(0, 0);
        io_flush();
        printf("[%d]\n", counter);
    }

    fixcursor();
    force_update = FALSE;
    return 0;
}

int VUTermPending()
{
    if(front != NULL) {
        if(VUGetAttr(front, BLOCKED)) {
            while(io_query())
                io_getc(0);
            VUTermAlert(1);
            return 0;
        }
    }
    else
        return 0;

    return io_query();
}

int VUTermStartup()
{
    char    *env;
    char    buf[32];
    int     fd;
    int     i;

    signal(SIGINT, SIG_IGN);

    for(i = 0; i < 256; ++i) {
        codepage[i] = (unsigned char)i;
        keymap[i] = (unsigned char)i;
    }

    env = getenv("VU_PROFILE");
    if(env)
        profile = atoi(env);

    env = getenv("VU_TRIM");
    if(env)
        trim = atoi(env);

    env = getenv("VU_COLORSET");
    if(!env)
        env = COLORSET;

    if((fd = open(env, O_RDONLY)) > 0) {
        read(fd, (char *)palette, sizeof(PALDEF) * 64);
        close(fd);
    }
    io_startup();

    env = getenv("BAUD");
    if(env) {
        if(atol(env) > 60000l)
            speed = 60000;
        else
            speed = atoi(env);
    }

    fd = -1;
    env = getenv("VU_TERMLIB");
    if(env)
        fd = open(env, O_RDONLY);

    if(fd > 0) {
        term = (TERMDEF *)malloc(sizeof(TERMDEF));
        read(fd, term, sizeof(TERMDEF));
        close(fd);
    }
    else {
        env = getenv("TERM");
        if(!env) {
            env = DEFTERM;
            strcpy(buf, "TERM=");
            strcat(buf, env);
            putenv(buf);
        }

        if(NULL == (term = LoadVUTermlib(AliasVUTermlib(env))))
            return 0;
    }
    rows = atoi(term->term_rows);
    cols = atoi(term->term_cols);
    io_escape(term->term_startup);
    io_flush();

    env = getenv("LINES");
    if(env && atoi(env))
        rows = atoi(env);

    env = getenv("COLUMNS");
    if(env && atoi(env))
        cols = atoi(env);

    if(!term->term_color_beg[0])
        color = FALSE;

    if(speed < 1200)
        color = FALSE;

    env = getenv("VU_COLOR");
    if(env)
        if(!stricmp("OFF", env))
            color = FALSE;

    viewport.tp = 0;
    viewport.lt = 0;
    viewport.rt = cols - 1;
    viewport.bt = rows - 2;
    eolmap = (unsigned char *)malloc(rows);
    if(NULL == (imageport = (CELL *)malloc(rows * cols * sizeof(CELL) * 2)))
        return 0;

    shadow_offset = rows * cols;
    for(i = 0; i < shadow_offset * 2; ++i)
        *(imageport + i) = 0x0000;          /* initialize in shadow */

    if(color || term->term_attr_inverse_on[0])
        inverse_mode = TRUE;

    if(!color) {
        term->term_color_beg[0] = 0;
        term->term_color_attr = FALSE;
        attrmask = MONOATTRMASK;
    }

    if(!term->term_attr_blink_on[0])
        attrmask &= ~BLINK;

    if(!term->term_attr_inverse_on[0])
        attrmask &= ~SUBINVERSE;

    if(!term->term_attr_bright_on[0])
        attrmask &= ~SUBBRIGHT;

    if(!term->term_attr_underline_on[0])
        attrmask &= ~SUBUNDERLINE;

    alt_base = atoi(term->key_alt_base);
    offset = atoi(term->term_cursor_offset);
    fcolor = (unsigned char)atoi(term->term_color_fore);
    bcolor = (unsigned char)atoi(term->term_color_back);
    if(term->char_shade)
        deskpattern = 0x0100 | GRAPHIC_CELL | BOX_SHADED;

    if(!term->term_attr_merge) {
        if(!term->term_attr_bright_off[0])
            clrmask |= BRIGHT;

        if(!term->term_attr_inverse_off[0])
            clrmask |= INVERSE;

        if(!term->term_attr_underline_off[0])
            clrmask |= UNDERLINE;

        if(!term->term_attr_blink_off[0])
            clrmask |= BLINK;
    }

    if(term->term_attr_graphics_on[0])
        graphics_mode = TRUE;

    tstmask = attrmask | BRIGHT;
    if(!term->term_attr_inverse_on[0])
        tstmask &= ~INVERSE;

    if(!term->term_attr_bright_on[0])
        tstmask &= ~BRIGHT;

    if(!term->term_attr_underline_on[0])
        tstmask &= ~UNDERLINE;

    if(term->term_color_attr)
        attributes |= 0x77;

    if(term->term_stat_line[0] || term->term_wrap_flag)
        inverse_status = TRUE;

    if(!color || !term->char_top)
        shadow_button = FALSE;

    return TRUE;
}


void    VUTermShutdown()
{
    setattr(DARKWHITE | BLACK);
    cursor(rows - 1, 0);
    io_escape(term->term_cursor_show);
    io_puts(term->term_eol);
    io_escape(term->term_shutdown);
    io_flush();
    free(imageport);
    free(eolmap);
    io_shutdown();
}

void    GetVUCodepage(int cp)
{
    char    path[128];
    int     fd;

    strcpy(path, CODEPAGES);
    strcat(path, term->term_id);
    fd = open(path, O_RDONLY);
    if(fd < 0)
        return;

    codepage_loaded = TRUE;
    lseek(fd, cp * sizeof(codepage), SEEK_SET);
    read(fd, codepage, sizeof(codepage));
    read(fd, path, 32);
    if(path[0]) /* init */
        write(0, path, strlen(path));
    read(fd, keymap + 32, sizeof(keymap) - 32);
    close(fd);
}









