/* Copyright (C) 1992-2014 David Sugar, Tycho Softworks.
 *
 * This file is part of worldvu.
 *
 * worldvu is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * worldvuvu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with worldvu.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <vu-config.h>
#include <worldvu/vu.h>
#include <worldvu/term.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

FIXEDWINDOW(Error);
FIXEDWINDOW(STTY);
FIXEDWINDOW(TermList);
FIXEDWINDOW(TermUpdate);
FIXEDWINDOW(Attrib);
FIXEDWINDOW(Cursor);
FIXEDWINDOW(Screen);
FIXEDWINDOW(Graphics);
FIXEDWINDOW(Keyboard);

enum {
    ID_INSERT=VU_USER,
    ID_UPDATE,
    ID_DELETE,
    ID_ID,
    ID_NAME,
    ID_ROWS,
    ID_COLS,
    ID_GOTO,
    ID_ATTR,
    ID_TERM,
    ID_GRAPHICS,
    ID_KEYBOARD,
    ID_EXPORT,
    ID_STTY,
    ID_DEFAULT,
    C_BEG=2000,
    C_SEP,
    C_END,
    C_VT,
    C_MODE,
    C_OFFSET,
    C_HIDE,
    C_SHOW,
    C_BIG,
    A_BEG=2100,
    A_END,
    A_SEP,
    A_MERGE,
    A_CLEAR,
    A_INV_ON,
    A_INV_OFF,
    A_BOLD_ON,
    A_BOLD_OFF,
    A_ULINE_ON,
    A_ULINE_OFF,
    A_GR_ON,
    A_GR_OFF,
    A_COLOR_BEG,
    A_COLOR_SEP,
    A_COLOR_END,
    A_COLOR_FORE,
    A_COLOR_BACK,
    A_COLOR_MODE,
    A_COLOR_ANSI,
    T_START=2200,
    T_STOP,
    T_BEEP,
    T_NEWLINE,
    T_CLS,
    T_EOL,
    T_EOP,
    T_LEFT,
    T_RIGHT,
    T_UP,
    T_DOWN,
    S_BREAK = 2300,
    S_QUIT,
    S_START,
    S_STOP,
    S_EOF,
    S_KILL,
    S_ERASE	
};

static	TERMDEF	term, defterm;
static	char	shade[4], solid[4], filltop[4], fillbot[4];
static	char	cards[4][4], arrows[4][4], boxes[16][4];

static	char	stty_break[4];
static	char	stty_quit[4];
static	char	stty_start[4];
static	char	stty_stop[4];
static	char	stty_eof[4];
static	char	stty_kill[4];
static	char	stty_erase[4];
static	char	init_startup[25];
static	char	init_shutdown[25];

CONST	char	digits[16] = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'};

PROGRAM
{
    char	*title = argv[1];
    TERMDEF	*entry;

    entry = LoadVUTermlib(AliasVUTermlib(getenv("TERM")));
    if(!entry) {
	fprintf(stderr, "*** %s: not found\n", getenv("TERM"));
	exit(-1);
    }
    memcpy(&term, entry, sizeof(term));
    memcpy(&defterm, entry, sizeof(term));

    if(!title)
	title = "Customize Terminal";

    VUStartup();
    VUCreateQ(VUDesktop, (STRING)title, -1, -1, NULL); 
    VUCreateQ(TermUpdate, NULL, -1, -1, NULL);

    while(VUDispatch())
	;

    VUShutdown();
    return 0;
}

static	void getstr(char *s)
{
    char	buf[32];
    char *	old = s;
    int		cnt = 0;

    while(*s) {
	if(*s == 0x1e) {
	    buf[cnt++] = '$';
	    buf[cnt++] = '1';
	    buf[cnt++] = 'e';
	}
	else if(*s < 32) {
	    buf[cnt++] = '^';
	    buf[cnt++] = (unsigned char)(*s + '@');
	}
	else if(*s > 127 && *s < 160) {
	    buf[cnt++] = '#';
	    buf[cnt++] = (unsigned char)((*s & 0x1f) + '@');
	}
	else if(*s == '#') {
	    buf[cnt++] = '#';
	    buf[cnt++] = '#';
	}
	else if(*s == '^') {
	    buf[cnt++] = '^';
	    buf[cnt++] = '^';
	}
	else if(*s == '$') {
	    buf[cnt++] = '$';
	    buf[cnt++] = '$';
	}
	else if(*s > 126) {
	    buf[cnt++] = '$'; 
	    buf[cnt++] = digits[*s / 16];
	    buf[cnt++] = digits[*s % 16];
	}
	else
	    buf[cnt++] = *s;
	++s;
    }
    buf[cnt] = 0;
    strcpy(old, buf);
    return;
}

static	void	putstr(s)
STRING s;
{
    int		cnt = 0;
    BOOL	cflag = FALSE;
    BOOL	hflag = FALSE;
    int		hexdigit = 0;
    int		hex = 0;
    STRING	buf = s;

    while(*s) {
	switch(*s) {
	case '$':
	    if(hexdigit) {
		hexdigit = 0;
		buf[cnt++] = '$';
	    
	    }
	    else {
		hexdigit = 1;
		hex = 0;
	    }
	    break;
	case '^':
	    if(!cflag)
		cflag = TRUE;
	    else {
		cflag = FALSE;
		buf[cnt++] = '^';
	    }
	    break;
	case '#':
	    if(!hflag)
		hflag = TRUE;
	    else {
		hflag = FALSE;
		buf[cnt++] = '#';
	    }
	    break;
	default:
	    if(hexdigit) {
		hex *= 16;
		if(*s >= '0' && *s <= '9')
		    hex += (*s - '0');
		else
		    hex += (tolower(*s) - 'a' + 10);
		if(++hexdigit > 2) {
		    buf[cnt++] = (unsigned char)hex;
		    hexdigit = 0;
		}
	    }
	    else if(hflag) {
		buf[cnt++] = (unsigned char)((*s & 0x1f) | 0x80);
		hflag = FALSE;
	    }
	    else if(cflag) {
		buf[cnt++] = (unsigned char)(*s & 0x1f);
		cflag = FALSE;
	    }
	    else
		buf[cnt++] = (unsigned char)*s;
	}
	++s;
    }
    buf[cnt] = 0;
    return;
}

static	void	expand_term()
{
    int	i;

    sprintf(shade, "%u", term.char_shade & 0xff);
    sprintf(solid, "%u", term.char_solid & 0xff);
    sprintf(filltop, "%u", term.char_top & 0xff);
    sprintf(fillbot, "%u", term.char_bot & 0xff);

    sprintf(stty_break, "%u", term.char_stty_break);
    sprintf(stty_quit, "%u", term.char_stty_quit);
    sprintf(stty_start, "%u", term.char_stty_start);
    sprintf(stty_stop, "%u", term.char_stty_stop);
    sprintf(stty_eof, "%u", term.char_stty_eof);
    sprintf(stty_kill, "%u", term.char_stty_kill);
    sprintf(stty_erase, "%u", term.char_stty_erase);
    
    for(i = 0; i < 4; ++i) {
	sprintf(cards[i], "%u", term.char_cards[i] & 0xff);
	sprintf(arrows[i], "%u", term.char_arrows[i] & 0xff);
	getstr(term.key_move[i]);
    }
    for(i = 0; i < 16; ++i)
	sprintf(boxes[i], "%u", term.char_boxset[i] & 0xff);

    getstr(term.term_cursor_beg);
    getstr(term.term_cursor_sep);
    getstr(term.term_cursor_end);
    getstr(term.term_cursor_vt);
    getstr(term.term_cursor_show);
    getstr(term.term_cursor_hide);
    getstr(term.term_cursor_big);
    getstr(term.term_attr_beg);
    getstr(term.term_attr_sep);
    getstr(term.term_attr_end);
    getstr(term.term_attr_clear);
    getstr(term.term_attr_inverse_on);
    getstr(term.term_attr_inverse_off);
    getstr(term.term_attr_bright_on);
    getstr(term.term_attr_bright_off);
    getstr(term.term_attr_underline_on);
    getstr(term.term_attr_underline_off);
    getstr(term.term_attr_graphics_on);
    getstr(term.term_attr_graphics_off);
    getstr(term.term_attr_blink_on);
    getstr(term.term_attr_blink_off);
    getstr(term.term_color_beg);
    getstr(term.term_color_sep);
    getstr(term.term_color_end);
    strcpy(init_startup, term.term_startup);
    getstr(init_startup);
    strcpy(init_shutdown, term.term_shutdown);
    getstr(init_shutdown);
    getstr(term.term_beep);
    getstr(term.term_cls);
    getstr(term.term_eol);
    getstr(term.term_eop);
    getstr(term.term_left);
    getstr(term.term_right);
    getstr(term.term_up);
    getstr(term.term_down);
    getstr(term.term_stat_line);
    getstr(term.term_window_beg);
    getstr(term.term_window_sep);
    getstr(term.term_window_end);
    getstr(term.term_window_clr);
    getstr(term.term_window_up);
    getstr(term.term_window_down);
    getstr(term.key_alt_prefix);
    for(i = 0; i < 20; ++i)
	getstr(term.key_table[i]);
}

static	void	compile()
{
    int	i;

    term.char_shade = (char)atoi(shade);
    term.char_solid = (char)atoi(solid);
term.char_top = (char)atoi(filltop);
    term.char_bot = (char)atoi(fillbot);
    
    for(i = 0; i < 4; ++i) {
	term.char_cards[i] = (char)atoi(cards[i]);
	term.char_arrows[i] = (char)atoi(arrows[i]);
	putstr(term.key_move[i]);
    }

    for(i = 0; i < 16; ++i)
	term.char_boxset[i] = (char)atoi(boxes[i]);

    term.char_stty_break = (unsigned char)atoi(stty_break);
    term.char_stty_quit = (unsigned char)atoi(stty_quit);
    term.char_stty_start = (unsigned char)atoi(stty_start);
    term.char_stty_stop = (unsigned char)atoi(stty_stop);
    term.char_stty_eof = (unsigned char)atoi(stty_eof);
    term.char_stty_kill = (unsigned char)atoi(stty_kill);
    term.char_stty_erase = (unsigned char)atoi(stty_erase);

    putstr(term.term_cursor_beg);
    putstr(term.term_cursor_sep);
    putstr(term.term_cursor_end);
    putstr(term.term_cursor_vt);
    putstr(term.term_cursor_show);
    putstr(term.term_cursor_hide);
    putstr(term.term_cursor_big);
    putstr(term.term_attr_beg);
    putstr(term.term_attr_sep);
    putstr(term.term_attr_end);
    putstr(term.term_attr_clear);
    putstr(term.term_attr_inverse_on);
    putstr(term.term_attr_inverse_off);
    putstr(term.term_attr_bright_on);
    putstr(term.term_attr_bright_off);
    putstr(term.term_attr_underline_on);
    putstr(term.term_attr_underline_off);
    putstr(term.term_attr_graphics_on);
    putstr(term.term_attr_graphics_off);
    putstr(term.term_attr_blink_on);
    putstr(term.term_attr_blink_off);
    putstr(term.term_color_beg);
    putstr(term.term_color_sep);
    putstr(term.term_color_end);
    putstr(init_startup);
    strcpy(term.term_startup, init_startup);
    putstr(init_shutdown);
    strcpy(term.term_shutdown, init_shutdown);
    putstr(term.term_beep);
    putstr(term.term_cls);
    putstr(term.term_eol);
    putstr(term.term_eop);
    putstr(term.term_left);
    putstr(term.term_right);
    putstr(term.term_up);
    putstr(term.term_down);
    putstr(term.term_stat_line);
    putstr(term.term_window_beg);
    putstr(term.term_window_end);
    putstr(term.term_window_sep);
    putstr(term.term_window_up);
    putstr(term.term_window_down);
    putstr(term.key_alt_prefix);
    for(i = 0; i < 20; ++i)
	putstr(term.key_table[i]);
}

DEFPORT(TermUpdatePort)
{
    int		rtn;
    int		fd;
    char	path[128];

    switch(event) {
    case VU_CREATE:
	snprintf(path, sizeof(path), "%s/.termdef", getenv("HOME"));
	fd = open(path, O_RDONLY);
	if(fd > 0) {
	    read(fd, &term, sizeof(term));
	    close(fd);
	}	
	expand_term();
	
	return USEPORT(VUFormPort);

    case VU_BUTTON:
	switch(p1) {
	case ID_DEFAULT:
	    memcpy(&term, &defterm, sizeof(term));
	    expand_term();
	    VUPostRedraw(VU_PAINT, vu, NULL);
	    return TRUE;
	case VU_ACCEPT:
	    rtn = USEPORT(VUFormPort);	
	    if(rtn) {
		compile();
		snprintf(path, sizeof(path), "%s/.termdef", getenv("HOME"));
		fd = open(path, O_WRONLY | O_CREAT, 0600);
		if(fd > 0) {
		    write(fd, &term, sizeof(term));
		    close(fd);
		}
	    }
	    return rtn;
	case ID_GOTO:
	    VUCreateQ(Cursor, NULL, -1, -1, NULL);
	    return TRUE;
	case ID_GRAPHICS:
	    VUCreateQ(Graphics, NULL, -1, -1, NULL);
	    return TRUE;
	case ID_KEYBOARD:
	    VUCreateQ(Keyboard, NULL, -1, -1, NULL);
	    return TRUE;
	case ID_TERM:
	    VUCreateQ(Screen, NULL, -1, -1, NULL);
	    return TRUE;
	case ID_ATTR:
	    VUCreateQ(Attrib, NULL, -1, -1, NULL);
	    return TRUE;
	case ID_STTY:
	    VUCreateQ(STTY, NULL, -1, -1, NULL);
	    return TRUE;
	}
    default:
	    break;
    }
    return USEPORT(VUFormPort);
}

const	DEFWINDOW(STTY)
    WIN_CLASS(VUFormPort, (STRING)"Terminal STTY Options")
    WIN_ATTR(13, 36, EDITBAR | HASSHADOW | HASMENUBAR | HASBORDER, 16,
	    (STRING)"Enter '0' for default or specify stty character code")

    TEXT_CONTROL((STRING)"(INTR)  ~Break character:",	2, 2, 1, 28, S_BREAK)
    EDITNUM_CONTROL(stty_break,                 2, 30, 3, S_BREAK, NULL)
    TEXT_CONTROL((STRING)"(QUIT)  ~Quit character:",    3, 2, 1, 28, S_QUIT)
    EDITNUM_CONTROL(stty_quit,                  3, 30, 3, S_QUIT, NULL)
    TEXT_CONTROL((STRING)"(START) ~Start character:",   4, 2, 1, 28, S_START)
    EDITNUM_CONTROL(stty_start,                 4, 30, 3, S_START, NULL)
    TEXT_CONTROL((STRING)"(STOP)  St~op character:",    5, 2, 1, 28, S_STOP)
    EDITNUM_CONTROL(stty_stop,                  5, 30, 3, S_STOP, NULL)
    TEXT_CONTROL("(EOF)   End of ~file char:",  6, 2, 1, 28, S_EOF)
    EDITNUM_CONTROL(stty_eof,                   6, 30, 3, S_EOF, NULL)
    TEXT_CONTROL("(KILL)  ~Kill line char:",    7, 2, 1, 28, S_KILL)
    EDITNUM_CONTROL(stty_kill,                  7, 30, 3, S_KILL, NULL)
    TEXT_CONTROL("(ERASE) ~Erase character:",   8, 2, 1, 28, S_ERASE)
    EDITNUM_CONTROL(stty_erase,                 8, 30, 3, S_ERASE, NULL)
    BUTTON_CONTROL(" ~Accept ", 	10, 14, 8, VU_ACCEPT)
ENDDEF

const	DEFWINDOW(Cursor)
    WIN_CLASS(VUFormPort, "Cursor Motion and Properties")
    WIN_ATTR(17, 32, EDITBAR | HASSHADOW | HASMENUBAR | HASBORDER, 16, 
	    "Complete Form")

    TEXT_CONTROL("~Begin cursor motion:",		2, 2, 1, 22, C_BEG) 
    EDITLN_CONTROL(term.term_cursor_beg,        2,24, 3, C_BEG, NULL)
    TEXT_CONTROL("Motion seperator:",           3, 2, 1, 22, C_SEP)
    EDITLN_CONTROL(term.term_cursor_sep,        3,24, 3, C_SEP, NULL)
    TEXT_CONTROL("End cursor motion:",          4, 2, 1, 22, C_END)
    EDITLN_CONTROL(term.term_cursor_end,        4,24, 3, C_END, NULL)
    TEXT_CONTROL("Vertical left motion:",       5, 2, 1, 22, C_VT)
    EDITLN_CONTROL(term.term_cursor_vt,         5,24, 3, C_VT, NULL)

    TEXT_CONTROL("Decimal motion values:",      7, 2, 1, 22, C_MODE)
    CHECK_CONTROL((char *)&term.term_cursor_decimal, 7, 24, C_MODE)
    TEXT_CONTROL("Motion offset by:",           8, 2, 1, 22, C_OFFSET)
    EDITNUM_CONTROL(term.term_cursor_offset,    8,24, 2, C_OFFSET, NULL)
    TEXT_CONTROL("Specify row first:",          9, 2, 1, 22, C_OFFSET)
    CHECK_CONTROL(&term.term_cursor_order,      9,24, C_OFFSET)

    TEXT_CONTROL("Hide cursor command:",       11, 2, 1, 22, C_HIDE)
    EDITLN_CONTROL(term.term_cursor_hide,      11,24, 6, C_HIDE, NULL)
    TEXT_CONTROL("Show cursor command:",       12, 2, 1, 22, C_SHOW)
    EDITLN_CONTROL(term.term_cursor_show,      12,24, 6, C_SHOW, NULL)
    TEXT_CONTROL("Insert cursor command:",     13, 2, 1, 22, C_BIG)
    EDITLN_CONTROL(term.term_cursor_big,       13,24, 6, C_BIG, NULL)

    BUTTON_CONTROL(" ~Accept ", 15, 12, 8, VU_ACCEPT)
ENDDEF

const	DEFWINDOW(Keyboard)
    WIN_CLASS(VUFormPort, "Keyboard table entries")
    WIN_ATTR(21, 46, EDITBAR | HASSHADOW | HASMENUBAR | HASBORDER, 16, "Complete Form")

    TEXT_CONTROL("F1-Key:",						2, 2, 1, 12, ID_KEYBOARD) 
    EDITLN_CONTROL(term.key_table[0],           2,14, 6, ID_KEYBOARD, NULL)
    TEXT_CONTROL("F2-Key:",					 	3, 2, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[1],           3,14, 6, 0, NULL)
    TEXT_CONTROL("F3-Key:",						4, 2, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[2],           4,14, 6, 0, NULL)
    TEXT_CONTROL("F4-Key:",						5, 2, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[3],           5,14, 6, 0, NULL)
    TEXT_CONTROL("F5-Key:",						6, 2, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[4],           6,14, 6, 0, NULL)
    TEXT_CONTROL("F6-Key:",						7, 2, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[5],           7,14, 6, 0, NULL)
    TEXT_CONTROL("F7-Key:",						8, 2, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[6],           8,14, 6, 0, NULL)
    TEXT_CONTROL("F8-Key:",						9, 2, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[7],           9,14, 6, 0, NULL)
    TEXT_CONTROL("F9-Key:",					   10, 2, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[8],          10,14, 6, 0, NULL)
    TEXT_CONTROL("F10-Key:",				   11, 2, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[9],          11,14, 6, 0, NULL)

    TEXT_CONTROL("Select-Key:",					2,22, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[10],          2,34, 6, 0, NULL)
    TEXT_CONTROL("Cancel-Key:",				 	3,22, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[11],          3,34, 6, 0, NULL)
    TEXT_CONTROL("Home-Key:",					4,22, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[12],          4,34, 6, 0, NULL)
    TEXT_CONTROL("End-Key:",					5,22, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[13],          5,34, 6, 0, NULL)
    TEXT_CONTROL("Page-up:",					6,22, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[14],          6,34, 6, 0, NULL)
    TEXT_CONTROL("Page-down:",					7,22, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[15],          7,34, 6, 0, NULL)
    TEXT_CONTROL("Insert:",				        8,22, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[16],          8,34, 6, 0, NULL)
    TEXT_CONTROL("Delete:",						9,22, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[17],          9,34, 6, 0, NULL)
    TEXT_CONTROL("Help-Key:",				   10,22, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[18],         10,34, 6, 0, NULL)
    TEXT_CONTROL("Find-Key:",				   11,22, 1, 12, 0) 
    EDITLN_CONTROL(term.key_table[19],         11,34, 6, 0, NULL)
    TEXT_CONTROL("Key-up:",                    12,22, 1, 12, 0)
    EDITLN_CONTROL(term.key_move[0],           12,34, 6, 0, NULL)
    TEXT_CONTROL("Key-down:",                  13,22, 1, 12, 0)
    EDITLN_CONTROL(term.key_move[1],           13,34, 6, 0, NULL)
    TEXT_CONTROL("Key-left:",                  14,22, 1, 12, 0)
    EDITLN_CONTROL(term.key_move[2],           14,34, 6, 0, NULL)
    TEXT_CONTROL("Key-right:",                 15,22, 1, 12, 0)
    EDITLN_CONTROL(term.key_move[3],           15,34, 6, 0, NULL)

    TEXT_CONTROL("Scancodes:",                 13, 2, 1, 12, 0)
    CHECK_CONTROL((char *)&term.key_scancodes, 13, 14, 0)
    TEXT_CONTROL("DEC Keypad:",                   14, 2, 1, 12, 0)
    CHECK_CONTROL((char *)&term.term_ansi_keypad, 14, 14, 0)
    TEXT_CONTROL("Alt base:",                  15, 2, 1, 12, 0) 
    EDITNUM_CONTROL(term.key_alt_base,         15,14,3, 0, NULL)
    TEXT_CONTROL("Alt prefix:",                16, 2, 1, 12, 0)
    EDITLN_CONTROL(term.key_alt_prefix,        16,14, 6, 0, NULL)

    BUTTON_CONTROL(" ~Accept ", 18, 19, 8, VU_ACCEPT)
ENDDEF

const	DEFWINDOW(Graphics)
    WIN_CLASS(VUFormPort, "Graphic characters and codes")
    WIN_ATTR(22, 38, EDITBAR | HASSHADOW | HASMENUBAR | HASBORDER, 16, "Complete Form")

    TEXT_CONTROL("Shaded box:",		2, 2, 1, 12, ID_GRAPHICS) 
    EDITNUM_CONTROL(shade,          2,14, 3, ID_GRAPHICS, NULL)
    TEXT_CONTROL("Solid box:",		3, 2, 1, 12, 0) 
    EDITNUM_CONTROL(solid,          3,14, 3, 0, NULL)
    TEXT_CONTROL("Top box:",		4, 2, 1, 12, 0) 
    EDITNUM_CONTROL(filltop,        4,14, 3, 0, NULL)
    TEXT_CONTROL("Bottom box:",		5, 2, 1, 12, 0) 
    EDITNUM_CONTROL(fillbot,        5,14, 3, 0, NULL)

    TEXT_CONTROL("Arrow up:",		7, 2, 1, 12, 0) 
    EDITNUM_CONTROL(arrows[0],      7,14, 3, 0, NULL)
    TEXT_CONTROL("Arrow down:",		8, 2, 1, 12, 0) 
    EDITNUM_CONTROL(arrows[1],      8,14, 3, 0, NULL)
    TEXT_CONTROL("Arrow left:",		9, 2, 1, 12, 0) 
    EDITNUM_CONTROL(arrows[2],      9,14, 3, 0, NULL)
    TEXT_CONTROL("Arrow right:",   10, 2, 1, 12, 0) 
    EDITNUM_CONTROL(arrows[3],     10,14, 3, 0, NULL)

    TEXT_CONTROL("Clubs:",	       12, 2, 1, 12, 0) 
    EDITNUM_CONTROL(cards[0],      12,14, 3, 0, NULL)
    TEXT_CONTROL("Diamonds:",	   13, 2, 1, 12, 0) 
    EDITNUM_CONTROL(cards[1],      13,14, 3, 0, NULL)
    TEXT_CONTROL("Hearts:",		   14, 2, 1, 12, 0) 
    EDITNUM_CONTROL(cards[2],      14,14, 3, 0, NULL)
    TEXT_CONTROL("Spades:",        15, 2, 1, 12, 0) 
    EDITNUM_CONTROL(cards[3],      15,14, 3, 0, NULL)

    
    TEXT_CONTROL("Empty cell:",		2,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[0],       2,32, 3, 0, NULL)
    TEXT_CONTROL("Up line:",	    3,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[1],       3,32, 3, 0, NULL)
    TEXT_CONTROL("Down line:",	    4,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[2],       4,32, 3, 0, NULL)
    TEXT_CONTROL("Ver. line:",	    5,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[3],       5,32, 3, 0, NULL)
    TEXT_CONTROL("Left line:",	    6,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[4],       6,32, 3, 0, NULL)
    TEXT_CONTROL("Left-up:",	    7,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[5],       7,32, 3, 0, NULL)
    TEXT_CONTROL("Left-down:",	    8,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[6],       8,32, 3, 0, NULL)
    TEXT_CONTROL("Left-tee:",	    9,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[7],       9,32, 3, 0, NULL)
    TEXT_CONTROL("Right line:",	   10,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[8],      10,32, 3, 0, NULL)
    TEXT_CONTROL("Right-up:",	   11,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[9],      11,32, 3, 0, NULL)
    TEXT_CONTROL("Right-down:",	   12,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[10],     12,32, 3, 0, NULL)
    TEXT_CONTROL("Right-tee:",	   13,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[11],     13,32, 3, 0, NULL)
    TEXT_CONTROL("Hor. line:",	   14,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[12],     14,32, 3, 0, NULL)
    TEXT_CONTROL("Up-tee:",	   	   15,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[13],     15,32, 3, 0, NULL)
    TEXT_CONTROL("Down-tee:",	   16,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[14],     16,32, 3, 0, NULL)
    TEXT_CONTROL("Cross:",		   17,20, 1, 12, 0) 
    EDITNUM_CONTROL(boxes[15],     17,32, 3, 0, NULL)

    BUTTON_CONTROL(" ~Accept ", 19, 15, 8, VU_ACCEPT)
ENDDEF

const	DEFWINDOW(Screen)
    WIN_CLASS(VUFormPort, "Screen Motion and Properties")
    WIN_ATTR(20, 50, EDITBAR | HASSHADOW | HASMENUBAR | HASBORDER, 16, "Complete Form")

    TEXT_CONTROL("I~nitialize terminal:",		2, 2, 1, 22, T_START) 
    EDITLN_CONTROL(init_startup,           2,24,24, T_START, NULL)
    TEXT_CONTROL("Restore terminal:",           3, 2, 1, 22, T_STOP)
    EDITLN_CONTROL(init_shutdown,          3,24,24, T_STOP, NULL)
    TEXT_CONTROL("Beep terminal:",              4, 2, 1, 22, T_BEEP)
    EDITLN_CONTROL(term.term_beep,              4,24, 6, T_BEEP, NULL)
    TEXT_CONTROL("Clear screen:",               5, 2, 1, 22, T_CLS)
    EDITLN_CONTROL(term.term_cls,               5,24, 6, T_CLS, NULL)
    TEXT_CONTROL("Clear end of page:",          6, 2, 1, 22, T_EOP)
    EDITLN_CONTROL(term.term_eop,               6,24, 6, T_EOP, NULL)
    TEXT_CONTROL("Clear end of line:",          7, 2, 1, 22, T_EOL)
    EDITLN_CONTROL(term.term_eol,               7,24, 6, T_EOL, NULL)
    TEXT_CONTROL("Fill status line:",           8, 2, 1, 22, 0)
    EDITLN_CONTROL(term.term_stat_line,         8,24, 16, 0, NULL)
    TEXT_CONTROL("No wrap at eol:",             9, 2, 1, 22, 0)  
    CHECK_CONTROL(&term.term_wrap_flag,         9,24, 0)
    TEXT_CONTROL("Unicode Encoding:",          10, 2, 1, 22, 0)  
    CHECK_CONTROL(&term.unicode_encoding,      10,24, 0)


    TEXT_CONTROL("Move cursor left:",          12, 2, 1, 22, T_LEFT)
    EDITLN_CONTROL(term.term_left,             12,24, 6, T_LEFT, NULL)
    TEXT_CONTROL("Move cursor right:",         13, 2, 1, 22, T_RIGHT)
    EDITLN_CONTROL(term.term_right,            13,24, 6, T_RIGHT, NULL)
    TEXT_CONTROL("Move cursor up:",            14, 2, 1, 22, T_UP)
    EDITLN_CONTROL(term.term_up,               14,24, 6, T_UP, NULL)
    TEXT_CONTROL("Move cursor down:",          15, 2, 1, 22, T_DOWN)
    EDITLN_CONTROL(term.term_down,             15,24, 6, T_DOWN, NULL)

    BUTTON_CONTROL(" ~Accept ", 17, 18, 8, VU_ACCEPT)
ENDDEF

const	DEFWINDOW(Attrib)
    WIN_CLASS(VUFormPort, "Text attribute control")
    WIN_ATTR(20, 64, EDITBAR | HASSHADOW | HASMENUBAR | HASBORDER, 16, "Complete Form")

    TEXT_CONTROL("~Begin attrib command:",		2, 2, 1, 22, A_BEG) 
    EDITLN_CONTROL(term.term_attr_beg,          2,24, 3, A_BEG, NULL)
    TEXT_CONTROL("Attribute seperator:",        3, 2, 1, 22, A_SEP)
    EDITLN_CONTROL(term.term_attr_sep,          3,24, 3, A_SEP, NULL)
    TEXT_CONTROL("End attrib command:",         4, 2, 1, 22, A_END)
    EDITLN_CONTROL(term.term_attr_end,          4,24, 3, A_END, NULL)
    TEXT_CONTROL("Merge attrib controls:",      5, 2, 1, 22, A_MERGE)
    CHECK_CONTROL((char *)&term.term_attr_merge,5, 24, A_MERGE)

    TEXT_CONTROL("Clear all attributes:",       7, 2, 1, 22, A_CLEAR)
    EDITLN_CONTROL(term.term_attr_clear,        7,24, 6, A_CLEAR, NULL)
    TEXT_CONTROL("Select Inverse Text:",        8, 2, 1, 22, A_INV_ON)
    EDITLN_CONTROL(term.term_attr_inverse_on,   8,24, 6, A_INV_ON, NULL)
    TEXT_CONTROL("Clear Inverse Text:",         9, 2, 1, 22, A_INV_OFF)
    EDITLN_CONTROL(term.term_attr_inverse_off,  9,24, 6, A_INV_OFF, NULL)
    TEXT_CONTROL("Select Bright Text:",        10, 2, 1, 22, A_BOLD_ON)
    EDITLN_CONTROL(term.term_attr_bright_on,   10,24, 6, A_BOLD_ON, NULL)
    TEXT_CONTROL("Clear Bright Text:",         11, 2, 1, 22, A_BOLD_OFF)
    EDITLN_CONTROL(term.term_attr_bright_off,  11,24, 6, A_BOLD_OFF, NULL)
    TEXT_CONTROL("Select Underline Text:",     12, 2, 1, 22, A_ULINE_ON)
    EDITLN_CONTROL(term.term_attr_underline_on,12,24, 6, A_ULINE_ON, NULL)
    TEXT_CONTROL("Clear Underline Text:",      13, 2, 1, 22, A_ULINE_OFF)
    EDITLN_CONTROL(term.term_attr_underline_off,13,24, 6, A_ULINE_OFF, NULL)
    TEXT_CONTROL("Blink mode on:",             14, 2, 1, 22, A_GR_ON)
    EDITLN_CONTROL(term.term_attr_blink_on,    14,24, 6, A_GR_ON, NULL)
    TEXT_CONTROL("Blink mode off:",            15, 2, 1, 22, A_GR_OFF)
    EDITLN_CONTROL(term.term_attr_blink_off,   15,24, 6, A_GR_OFF, NULL)

    TEXT_CONTROL("B~egin color command:",		2,34, 1, 22, A_COLOR_BEG) 
    EDITLN_CONTROL(term.term_color_beg,         2,56, 6, A_COLOR_BEG, NULL)
    TEXT_CONTROL("Color seperator:",            3,34, 1, 22, A_COLOR_SEP)
    EDITLN_CONTROL(term.term_color_sep,         3,56, 6, A_COLOR_SEP, NULL)
    TEXT_CONTROL("End color command:",          4,34, 1, 22, A_COLOR_END)
    EDITLN_CONTROL(term.term_color_end,         4,56, 6, A_COLOR_END, NULL)

    TEXT_CONTROL("Foreground offset:",          6,34, 1, 22, A_COLOR_FORE)
    EDITNUM_CONTROL(term.term_color_fore,       6,56, 3, A_COLOR_FORE, NULL)
    TEXT_CONTROL("Background offset:",          7,34, 1, 22, A_COLOR_BACK)
    EDITNUM_CONTROL(term.term_color_back,       7,56, 3, A_COLOR_BACK, NULL)
    TEXT_CONTROL("Forground first:",            8,34, 1, 22, A_COLOR_MODE)
    CHECK_CONTROL(&term.term_color_order,       8,56, A_COLOR_MODE)
    TEXT_CONTROL("Ansi color table:",           9,34, 1, 22, A_COLOR_MODE)
    CHECK_CONTROL(&term.term_color_ansi,        9,56, A_COLOR_MODE)
    TEXT_CONTROL("Color attr merge:",          10,34, 1, 22, A_COLOR_MODE)
    CHECK_CONTROL(&term.term_color_attr,       10,56, A_COLOR_MODE)

    TEXT_CONTROL("Graphics mode on:",          12,34, 1, 22, A_GR_ON)
    EDITLN_CONTROL(term.term_attr_graphics_on, 12,56, 6, A_GR_ON, NULL)
    TEXT_CONTROL("Graphics mode off:",         13,34, 1, 22, A_GR_OFF)
    EDITLN_CONTROL(term.term_attr_graphics_off,13,56, 6, A_GR_OFF, NULL)

    BUTTON_CONTROL(" ~Accept ", 17, 28, 8, VU_ACCEPT)
ENDDEF

const	DEFWINDOW(TermUpdate)
    WIN_CLASS(TermUpdatePort, (STRING)"Custom User Terminal Definition")
    WIN_ATTR(22, 64, EDITBAR | HASSHADOW | HASMENUBAR | HASBORDER | DESKTOP, 16, 
	    "Press <ESCAPE> to exit")

    TEXT_CONTROL("Export:   ",                  2, 28,1, 9, ID_EXPORT)  
    EDITLN_CONTROL(term.term_export_id,         2, 36,12, ID_EXPORT, "Must enter export identifier")

    TEXT_CONTROL("Name:     ",                  3, 2, 1, 11, ID_NAME)	
    EDITLN_CONTROL(term.term_name,              3,12, 32, ID_NAME, "Must enter terminal name")  

    TEXT_CONTROL("Rows:",                       2,50, 1, 6, ID_ROWS)
    EDITNUM_CONTROL(term.term_rows,             2,56, 3, ID_ROWS, "Must enter screen size")
    
    TEXT_CONTROL("Rows:",                       3,50, 1, 6, ID_COLS)
    EDITNUM_CONTROL(term.term_cols,             3,56, 3, ID_COLS, "Must enter screen size")

    BOX_CONTROL("Properties:",                  4, 2,14, 60, 0)  
    BUTTON_CONTROL(" A~ttrib ",					5, 4, 8, ID_ATTR)
    TEXT_CONTROL("Text attribute control",      5,16, 1, 32, ID_ATTR)
    BUTTON_CONTROL(" C~ursor ",					7, 4, 8, ID_GOTO)
    TEXT_CONTROL("Cursor motion and properties",7,16, 1, 32, ID_GOTO)
    BUTTON_CONTROL(" ~Graph  ",				    9, 4, 8, ID_GRAPHICS)
    TEXT_CONTROL("Graphic character codes",     9,16, 1, 32, ID_GRAPHICS)
    BUTTON_CONTROL("  ~Keys  ",				   11, 4, 8, ID_KEYBOARD)
    TEXT_CONTROL("Keyboard control codes",     11,16, 1, 32, ID_KEYBOARD)
    BUTTON_CONTROL(" ~Screen ",				   13, 4, 8, ID_TERM)
    TEXT_CONTROL("Screen motion and properties",13,16, 1, 32, ID_TERM)
    BUTTON_CONTROL("  STT~Y  ",                 15, 4, 8, ID_STTY)
    TEXT_CONTROL("STTY character settings",     15,16, 1, 32, ID_STTY)
    
    TEXT_CONTROL("Original:",                   2, 2, 1, 11, ID_ID)
    TEXT_CONTROL(term.term_id,                  2,12, 1, 10, ID_ID)

    BUTTON_CONTROL(" ~Accept ",                19, 8, 8, VU_ACCEPT)
    BUTTON_CONTROL("~Default ",                19,28, 8, ID_DEFAULT)
    BUTTON_CONTROL(" ~Cancel ",                19,48, 8, VU_CANCEL)
ENDDEF
	

	
